from __future__ import annotations

"""

Implements the MCTS + Self-Refine algorithm from
`Accessing GPT-4 level Mathematical Olympiad Solutions via Monte
Carlo Tree Self-refine with LLaMa-3 8B: A Technical Report`
by Zhang et. al.

The authors' [repo](https://github.com/trotsky1997/MathBlackBox) uses critiques,
refinements, and parent nodes' answers as conversation history.
I haven't tried it yet.

"""

"""

### MCTS任务构建

MCT 自优化算法代表了蒙特卡洛树搜索（MCTS）与大型语言模型的结合，将不同的场景执行任务过程抽象为搜索树结构。树上的节点代表当前不同视角下的选择策略，而边表示主体对自身的反思。该算法的操作流程遵循 MCTS 算法的一般模式。

具体来说，我们采用模型的先验知识，来让主体通过一系列的自身反思探索来获取自身对问题的最优答案；这种方式依赖模型的对齐偏好，我们在每种偏好上设计了一个工程框架，来完成自我对不同答案的奖励进行采样策略


1、对问题生成的子任务，生成一个合理的规划的节点
2、对每个节点创建一个MCTS任务，
3、输入 problem（总问题的子任务相关的子问题）
4、评分代码重构，将片段摘录器模块集成到一个关于_evaluate_answer逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于evaluate_system_prompt 的生成策略，具体的为编写一个关于带有评估的评估器，由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message

5、自我反思代码重构,将片段摘录器模块集成到一个关于self_refine逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于critic_system_prompt和refine_system_prompt的生成策略，critic_system_prompt为生成一个关于子问题相关的loader_cosplay_builder中自身不完美的评价内容，refine_system_prompt为不完美评价的思考过程和评分值。
具体的为编写一个关于带有评价的生成器和反思生成器，它们由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message



#### MCTS执行中的重要环节
MCTS中的约束规则如下，需要保证这些节点必须符合下面所定义的基本规则

提示约束：模型在奖励评分期间必须遵守最严格的标准。生成结果需要为JSON Response format
{
    "thought": "The thought process behind the answer.",
    "answer": "A float representing the answer to the problem."
}


高分抑制：评分节点中不存在满分反馈机制；任何超过 95 分的奖励都会按固定金额减少，以遏制过高分数。

重复采样：每次访问搜索树节点都涉及对节点奖励的重复采样，以增强自我评估的可靠性。需要注意的是，当对节点的子节点进行奖励采样时，我们也会对其父节点进行奖励采样，以增加奖励采样的样本量。

"""

import random
import math
from collections import deque
from enum import Enum 
from typing import Tuple
from pydantic import BaseModel
import tqdm
from dreamsboard.engine.memory.mctsr.prompt import (
    gpt_prompt_config,
    RefineResponse,
)
from dreamsboard.engine.task_engine_builder.core import CodeGeneratorBuilder
from dreamsboard.document_loaders.structured_storyboard_loader import LinkedListNode

from dreamsboard.engine.generate.code_generate import QueryProgramGenerator, BaseProgramGenerator
from langchain.schema import AIMessage
from dreamsboard.common.try_parse_json_object import try_parse_json_object
from dreamsboard.document_loaders.kor_loader import KorLoader
from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepRefineNode
from langchain.prompts import PromptTemplate
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from dreamsboard.engine.storage.storage_context import StorageContext
import numpy as np
import logging
import re
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

logger.addHandler(handler)


PATTERN = re.compile(r"thought:\s*(.*?)\nanswer:\s*([\d.]+)", re.DOTALL)

PLAINTEXT_PATTERN = re.compile(r"```plaintext?([\s\S]*?)```[\s\S]*?answer: (\d+(\.\d+)?)", re.DOTALL)

ROOT_UCT_SCORE = 10_000


class MCTSNode(BaseModel):
    base_path: str
    answer: str
    linked_list_node: LinkedListNode
    """当前任务的会话信息""" 
    storage_context: StorageContext
    """当前任务的会话存储""" 
    parent: MCTSNode | None = None
    children: list[MCTSNode] = []
    visits: int = 0
    Q: float = 0
    reward_samples: list[int] = []

    class Config:
        arbitrary_types_allowed = True

    def add_child(self, child_node: MCTSNode):
        self.children.append(child_node)

    def __repr__(self):
        return f"MCTSNode(answer={self.answer}, Q={self.Q:.2f}, visits={self.visits})"

    def add_reward(self, reward: int):
        self.reward_samples.append(reward)
        avg_reward = np.mean(self.reward_samples)
        min_reward = np.min(self.reward_samples)

        # Average worst-case and average outcomes
        self.Q = (min_reward + avg_reward) / 2


class SelectionPolicy(Enum):
    GREEDY = 1
    IMPORTANCE_SAMPLING = 2
    PAIRWISE_IMPORTANCE_SAMPLING = 3

 


class MCTSr(BaseModel):
    
    llm_runable: Runnable[LanguageModelInput, BaseMessage]
    problem: str
    max_rollouts: int
    exploration_constant: float = 1.0
    max_children: int = 2
    epsilon: float = 1e-10
    reward_limit: int = 95
    excess_reward_penalty: int = 5
    selection_policy: SelectionPolicy = SelectionPolicy.GREEDY 

    root: MCTSNode | None = None

    # Logs
    critiques: list[str] = []
    refinements: list[str] = []
    rewards: list[float] = []
    selected_nodes: list[MCTSNode] = []

    class Config:
        arbitrary_types_allowed = True
        
    def self_refine(self, node: MCTSNode) -> Tuple[MCTSNode, RefineResponse]:
        raise NotImplementedError()

    def _evaluate_answer(self, node: MCTSNode) -> int:
        raise NotImplementedError()

    def self_evaluate(self, node: MCTSNode):
        """Evaluate the quality of the answer. Sample `num_samples` times and average the results."""
        reward = self._evaluate_answer(node)

        if reward > self.reward_limit:
            reward -= self.excess_reward_penalty

        node.add_reward(reward)

    def backpropagate(self, node: MCTSNode):
        parent = node.parent
        while parent:
            if parent.children is None or len(parent.children) == 0:
                parent = parent.parent
                continue

            best_child_Q = max(child.Q for child in parent.children)
            parent.Q = (parent.Q + best_child_Q) / 2
            parent.visits += 1
            parent = parent.parent

    def uct(self, node: MCTSNode):
        if not node.parent:
            # Using an arbitrarily high UCT score for the root node.
            # helps to prioritize breadth.
            return ROOT_UCT_SCORE

        return node.Q + self.exploration_constant * math.sqrt(
            math.log(node.parent.visits + 1) / (node.visits + self.epsilon)
        )

    def is_fully_expanded(self, node: MCTSNode):
        return len(node.children) >= self.max_children or any(
            child.Q > node.Q for child in node.children
        )

    def select_node(self):
        """Select a non-fully expanded node with the highest UCT value.

        A node is fully expanded if either:
        1. It has reached the max number of children
        2. Any of its children have a Q value greater than its own
        """
        candidates: list[MCTSNode] = []
        to_consider = deque([self.root])

        while to_consider:
            current_node = to_consider.popleft()
            if not self.is_fully_expanded(current_node):
                candidates.append(current_node)
            to_consider.extend(current_node.children)

        if not candidates:
            return self.root

        if self.selection_policy == SelectionPolicy.GREEDY:
            return max(candidates, key=self.uct)
        elif self.selection_policy == SelectionPolicy.IMPORTANCE_SAMPLING:
            # Sample, weighted by UCT score
            uct_scores = [self.uct(node) for node in candidates]
            selected_pair_idx = random.choices(
                range(len(candidates)), weights=uct_scores, k=1
            )[0]
            return candidates[selected_pair_idx]
        elif self.selection_policy == SelectionPolicy.PAIRWISE_IMPORTANCE_SAMPLING:
            # Sample, weighted by the difference in UCT scores between pairs
            uct_scores = [self.uct(node) for node in candidates]
            pairs = [
                (i, j) for i in range(len(candidates)) for j in range(len(candidates))
            ]
            pair_weights = [
                max(uct_scores[i], uct_scores[j]) - min(uct_scores[i], uct_scores[j])
                for i, j in pairs
            ]
            selected_pair_idx = random.choices(
                range(len(pairs)), weights=pair_weights, k=1
            )[0]
            selected_candidate_idx = max(
                pairs[selected_pair_idx], key=lambda x: uct_scores[x]
            )
            return candidates[selected_candidate_idx]
        else:
            raise ValueError(f"Invalid selection policy: {self.selection_policy}")
 

    def initialize(self, root_node: MCTSNode):
        """Generate a zero-shot answer."""
        if not isinstance(root_node, MCTSNode):
            raise ValueError("root_node must be an instance of MCTSNode")
        self.root = root_node

    def run(self): 
        for _ in tqdm.tqdm(range(self.max_rollouts)):
            node = self.select_node()
            self.self_evaluate(node)
            child, refined_answer = self.self_refine(node)
            node.add_child(child)
          
            if refined_answer.answer_score > self.reward_limit:
                refined_answer.answer_score -= self.excess_reward_penalty

            child.add_reward(refined_answer.answer_score)
            self.backpropagate(child)

        return self.get_best_answer()

    def get_best_answer(self):
        from collections import deque

        to_visit = deque([self.root])
        best_node = self.root

        while to_visit:
            current_node = to_visit.popleft()
            if current_node.Q > best_node.Q:
                best_node = current_node
            to_visit.extend(current_node.children)

        return best_node.answer

    def print(self):
        print_tree(self.root)



class MCTSrStoryboard(MCTSr): 

    def self_refine(self, node: MCTSNode) ->  Tuple[MCTSNode, RefineResponse]:
        """
        自我反思
        """

        critic_system_prompt_template =  PromptTemplate(
            input_variables=[
                "problem", 
                "current_answer",
                "context",
                "past_steps",
                "start_task_context",
                "aemo_representation_context",
                "task_step_name",
                "task_step_description",
                "task_step_level"
            ],
            template=os.environ.get(
                "critic_system_prompt_data", gpt_prompt_config.critic_system_prompt_data 
            )
        )
        past_steps = ''
        past_context = ''
        context_linked_list_node = node.linked_list_node.head
        past_steps += f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n'
        past_context += f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_question_answer}\n'
        while context_linked_list_node is not None and context_linked_list_node.next is not None:
            if context_linked_list_node.task_step_id == node.linked_list_node.task_step_id:
                break
            context_linked_list_node = context_linked_list_node.next
            
            past_steps += f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n'
            past_context += f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_question_answer}\n'
            

        user_prompt = critic_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            context=past_context,
            past_steps=past_steps,
            start_task_context=node.linked_list_node.start_task_context,
            aemo_representation_context=node.linked_list_node.aemo_representation_context,
            task_step_name=node.linked_list_node.task_step_name,
            task_step_description=node.linked_list_node.task_step_description,
            task_step_level=node.linked_list_node.task_step_level
        ) 
        _ai_message = self._get_ai_message(gpt_prompt_config.critic_system_prompt, user_prompt, node)
        assert _ai_message.content is not None
        critique = _ai_message.content
        assert critique is not None
        self.critiques.append(critique)
   
        refine_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem", 
                "current_answer",
                "critique",
                "context",
                "past_steps",
            ],
            template=os.environ.get(
                "refine_system_prompt_data", gpt_prompt_config.refine_system_prompt_data
            )
        )
        refine_system_prompt = refine_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            critique=critique,
            context=past_context,
            past_steps=past_steps
        )
 
        _refined_answer_response_message = self._get_ai_message(gpt_prompt_config.refine_system_prompt, refine_system_prompt, node)
        assert _refined_answer_response_message.content is not None
            
        json_object = {}
        try:
            task_step_refine_node_list = self._kor_task_step_refine_builder(_refined_answer_response_message)
            # 将列表answer_score平均
            answer_score_list: list[float] = []
            for task_step_refine_node in task_step_refine_node_list:
                try:
                    answer_score_list.append(float(task_step_refine_node.answer_socre))
                except ValueError:
                    answer_score_list.append(0.0)
            answer_score = sum(answer_score_list) / len(answer_score_list)
            if answer_score <= 1:
                answer_score = answer_score * 100

            json_object.update({
                "thought": task_step_refine_node_list[0].thought,
                "answer": task_step_refine_node_list[0].answer,
                "answer_score": answer_score
            })
        except Exception as e: 
            json_object = {
                "thought": "解析失败",
                "answer": node.answer,
                "answer_score": 0
            }

        logger.info("\033[1;32m" + f"解析后的 JSON 对象: {json_object}" + "\033[0m")
        refined_answer = RefineResponse.model_validate(json_object)
        self.refinements.append(refined_answer)

        # ```thought \n{refined_answer.thought} ```\n\n ```answer_score \n{refined_answer.answer_score} ```
        return MCTSNode(
            base_path=node.base_path,
            answer=f"{refined_answer.answer}",
            linked_list_node=node.linked_list_node,
            storage_context=node.storage_context,
            parent=node,
            children=[],
            visits=0,
            Q=0,
            reward_samples=[]
        ), refined_answer

    def _evaluate_answer(self, node: MCTSNode) -> int:
        """
        评估答案
        """ 
 
        evaluate_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem", 
                "answer"
            ],
            template=os.environ.get(
                "evaluate_system_prompt_data", gpt_prompt_config.evaluate_system_prompt_data
            )
        )

        user_prompt = evaluate_system_prompt_template.format(
            problem=self.problem,
            answer=node.answer,
        )
        
        for attempt in range(3):
            try:
                _ai_message = self._get_ai_message(gpt_prompt_config.evaluate_system_prompt, user_prompt, node) 
                assert _ai_message.content is not None
                return int(_ai_message.content)
            except ValueError: 
                user_prompt = f"{_ai_message.content}\n\nFailed to parse reward as an integer."
               
                if attempt == 2:
                    raise
    
    def _get_ai_message(self, system_prompt: str, user_prompt: str, node: MCTSNode) -> AIMessage:
        """
        获取AI消息
        """
        
        code_gen_builder = CodeGeneratorBuilder.from_template(nodes=[], storage_context=node.storage_context)
        _base_render_data = {
            'system_prompt': system_prompt,
            'messages': [user_prompt]
        }
        code_gen_builder.add_generator(BaseProgramGenerator.from_config(cfg={
            "code_file": "base_template_system.py-tpl",
            "render_data": _base_render_data,
        }))
 
        executor = code_gen_builder.build_executor(
            llm_runable=self.llm_runable,
            messages=[]
        )
        executor.execute()
        _ai_message = executor.chat_run()

        logger.info("\033[1;32m" + f"_ai_message: {_ai_message}" + "\033[0m")
        assert executor._ai_message is not None 

        return _ai_message
    
    def _kor_task_step_refine_builder(self, refined_answer_response_message: AIMessage) -> list[TaskStepRefineNode]:
        """
        抽取根据批评意见优化当前回答并续写上下文内容
        """
        kor_task_step_refine_builder = KorLoader.form_kor_task_step_refine_builder(self.llm_runable)
        response = kor_task_step_refine_builder.run(refined_answer_response_message.content)
        task_step_refine_node_list = []
        if response.get('data') is not None and response.get('data').get('script') is not None:
            step_list = response.get('data').get('script')
            for step in step_list:
                task_step_refine_node = TaskStepRefineNode(
                                            thought=step.get('thought'),
                                            answer=step.get('answer'),
                                            answer_socre=step.get('answer_socre')
                                            )
                task_step_refine_node_list.append(task_step_refine_node)

        return task_step_refine_node_list

def print_tree(node: MCTSNode | None, level: int = 0):
    if node is None:
        return
    indent = " " * level * 2
    node_str = repr(node)
    for line in node_str.split("\n"):
        logger.info(indent + line)
    for child in node.children:
        print_tree(child, level + 1)


from weaviate.classes.init import AdditionalConfig, Timeout

from weaviate.client import WeaviateAsyncClient, WeaviateClient
from weaviate.classes.config import Configure, Property, DataType, VectorDistances
import weaviate

import os


def init_context_connect() -> WeaviateClient:

    headers = {
        "X-Zhipuai-Api-Key": os.environ.get("ZHIPUAI_API_KEY")
    }
    client = weaviate.connect_to_local(
        headers=headers,
        additional_config=AdditionalConfig(
            timeout=Timeout(init=30, query=12000, insert=120)  # Values in seconds
        )
    )
        
    print(client.is_ready())
    return client


def init_context_collections(client: WeaviateClient, collection_id: str) -> tuple[str, str]:
    from datetime import datetime
 

    # Collection name with timestamp
    collection_name = "AtomgitPapers"+ collection_id 
    
    collection_name_context = f"{collection_name}Context"
    client.collections.delete(collection_name) 
    client.collections.delete(collection_name_context) 


    client.collections.create(
        collection_name,
        reranker_config=Configure.Reranker.transformers(),
        generative_config=Configure.Generative.zhipuai(
            # These parameters are optional
            model="glm-4-plus",   
            max_tokens=500, 
            temperature=0.7,
            top_p=0.7
        ),
        vectorizer_config=[
            
            # Set another named vector
            Configure.NamedVectors.text2vec_transformers(   
                name="chunk_text", source_properties=["chunk_text"],  
                vector_index_config=Configure.VectorIndex.hnsw(
                    distance_metric=VectorDistances.COSINE
                ) 
            ), 
        ],
    
        properties=[  # Define properties
            Property(name="ref_id", data_type=DataType.TEXT),
            Property(name="paper_id", data_type=DataType.TEXT),
            Property(name="paper_title", data_type=DataType.TEXT),
            Property(name="chunk_id", data_type=DataType.NUMBER),
            Property(name="chunk_text", data_type=DataType.TEXT),
            Property(name="original_filename", data_type=DataType.TEXT)
            
        ],
    )


    client.collections.create(
        collection_name_context,
        
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary( vectorize_collection_name=False),
    
        properties=[  # Define properties
            Property(name="refId", data_type=DataType.TEXT),
            Property(name="paperId", data_type=DataType.TEXT),
            Property(name="chunkId", data_type=DataType.NUMBER),
            Property(name="chunkText", data_type=DataType.TEXT),
            
        ],
    )

    return collection_name, collection_name_context





"""
处理分镜结构化信息根据分镜画面，解析分镜中存在的信息、
输入为json结构数据
[{
time_point_len:00:00:01,
perv_story_board:""
story_board:"/story_board0.png"
next_story_board:"/story_board1.png"
},
{
time_point_len:00:00:05,
perv_story_board:"/story_board0.png"
story_board:"/story_board1.png"
next_story_board:"/story_board2.png"
}]

输出格式为
分镜编号	场景编号	时间点	描述	备注
Shot 1	story_board0	00:00:01	描述分镜细节 备注
Shot 2	story_board2	00:00:05	描述分镜细节 备注


"""
from abc import ABC
from typing import List, Dict, Any
from langchain.document_loaders.base import BaseLoader
from langchain.docstore.document import Document
import pandas as pd
from pandas import DataFrame

class QuestionContext:
    def __init__(self, ref_id: str, score: float, text: str):
        self.ref_id = ref_id
        self.score = score
        self.text = text

# 定义链表节点类
class LinkedListNode:
    def __init__(self, 
                task_step_id: str,
                shot_number: int, 
                scene_number: str, 
                start_task_context: str=None, 
                aemo_representation_context: str=None,
                task_step_name: str=None, 
                task_step_description: str=None, 
                task_step_level: str=None,
                task_step_question: str=None,
                task_step_question_context: List[QuestionContext]=None,
                task_step_question_answer: str=None,
                ref_task_step_id: str=None
        ):
        self.task_step_id = task_step_id
        self.shot_number = shot_number
        self.scene_number = scene_number
        self.start_task_context = start_task_context
        self.aemo_representation_context = aemo_representation_context
        self.task_step_name = task_step_name
        self.task_step_description = task_step_description
        self.task_step_level = task_step_level
        self.task_step_question = task_step_question
        self.task_step_question_context = task_step_question_context
        self.task_step_question_answer = task_step_question_answer
        self.ref_task_step_id = ref_task_step_id
        self.head: LinkedListNode = None
        self.prev: LinkedListNode = None
        self.next: LinkedListNode = None


class StructuredStoryboard:
    """
    解析json数组中整理好的镜头，实现链表结构的处理操作
    """
    head: LinkedListNode = None
    prev_node: LinkedListNode = None

    def __init__(self, json_data: List[Dict[str, Any]]):
        """Initialize with  json_data."""
        self._parse_dialogue(json_data)

    def _parse_dialogue(self, json_data: List[Dict[str, Any]]) -> None:
        """
        解析JSON数组并创建链表
        """
        if self.prev_node is None:
            shot_number = 1
        else:
            shot_number = self.prev_node.shot_number+1

        for index, data in enumerate(json_data):
            scene_number = "story_board" + str(index)
            task_step_id = data["id_"]
            start_task_context = data["start_task_context"]
            aemo_representation_context = data["aemo_representation_context"]
            task_step_name = data["task_step_name"]
            task_step_description = data["task_step_description"]
            task_step_level = data["task_step_level"]
            task_step_question = data["task_step_question"]
            task_step_question_context = [QuestionContext(ref_id=context["ref_id"], score=context["score"], text=context["text"]) for context in data["task_step_question_context"]]
            task_step_question_answer = data["task_step_question_answer"]
            ref_task_step_id = data["ref_task_step_id"]

            node = LinkedListNode(
                task_step_id, 
                shot_number,
                scene_number,
                start_task_context,
                aemo_representation_context, 
                task_step_name, 
                task_step_description, 
                task_step_level, 
                task_step_question, 
                task_step_question_context, 
                task_step_question_answer, 
                ref_task_step_id
            )
            if self.prev_node:
                self.prev_node.next = node
                node.prev = self.prev_node

            else:
                self.head = node

            node.head = self.head
            self.prev_node = node
            shot_number += 1

    def get_task_step_node(self, task_step_id: str) -> LinkedListNode:
        """
        获取任务步骤节点
        """
        current_parse_node = self.head
        while current_parse_node is not None:
            if current_parse_node.task_step_id == task_step_id:
                return current_parse_node
            current_parse_node = current_parse_node.next
        return None
    def parse_table(self) -> DataFrame:
        """
        输出格式为
            任务步骤编号	分镜编号	场景编号    开始任务	任务总体描述	任务步骤名称	任务步骤描述	任务步骤层级	任务步骤问题	任务步骤问题上下文	任务步骤问题答案	参考任务步骤编号
        :return: DataFrame
        """
        table_data = []
        current_parse_node = self.head
        while current_parse_node is not None:
            row = [
                current_parse_node.task_step_id,
                current_parse_node.shot_number,
                current_parse_node.scene_number,
                current_parse_node.start_task_context,
                current_parse_node.aemo_representation_context,
                current_parse_node.task_step_name,
                current_parse_node.task_step_description,
                current_parse_node.task_step_level,
                current_parse_node.task_step_question,
                current_parse_node.task_step_question_context,
                current_parse_node.task_step_question_answer,
                current_parse_node.ref_task_step_id
            ]
            table_data.append(row)
            current_parse_node = current_parse_node.next

        table = pd.DataFrame(table_data, columns=[
            "task_step_id", 
            "shot_number", 
            "scene_number", 
            "start_task_context", 
            "aemo_representation_context", 
            "task_step_name", 
            "task_step_description", 
            "task_step_level", 
            "task_step_question", 
            "task_step_question_context", 
            "task_step_question_answer", 
            "ref_task_step_id"
        ])
        return table


class StructuredStoryboardLoader(BaseLoader, ABC):
    structured_storyboard: StructuredStoryboard

    def __init__(self, structured_storyboard: StructuredStoryboard):
        """Initialize with  json_data."""
        self.structured_storyboard = structured_storyboard

    def load(self) -> List[Document]:
        documents = []

        table = self.structured_storyboard.parse_table()
        for row in table.itertuples():

            raw_transcript_with_meta_info = (
                f"task_step_id: {row.task_step_id},"
                f"scene_number: {row.scene_number},"
                f"shot_number: {row.shot_number}\n\n"
                f"start_task_context: {row.start_task_context}\n\n"
                f"aemo_representation_context: {row.aemo_representation_context}\n\n"
                f"task_step_name: {row.task_step_name}\n\n"
                f"task_step_description: {row.task_step_description}\n\n"
                f"task_step_level: {row.task_step_level}\n\n"
                f"task_step_question: {row.task_step_question}\n\n"
                f"task_step_question_context: {row.task_step_question_context}\n\n"
                f"task_step_question_answer: {row.task_step_question_answer}\n\n"
                f"ref_task_step_id: {row.ref_task_step_id}\n\n"
            )

            structured_storyboard_json = row.to_json(orient='records',  indent=4)
            doc = Document(page_content=raw_transcript_with_meta_info, metadata=structured_storyboard_json)

            documents.append(doc)

        return documents





from dreamsboard.common.registry import registry
import os


root_dir = os.path.dirname(os.path.abspath(__file__))
registry.register_path("templates_library_root", root_dir)


def get_template_path(name):
    return os.path.join(registry.get_path("templates_library_root"), name)



from __future__ import annotations
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from abc import ABC
from typing import Any, Dict
from langchain.chains import LLMChain
from langchain.chains.base import Chain
from langchain_core.prompts  import PromptTemplate
from langchain.chains import SequentialChain
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableParallel, RunnableLambda

from dreamsboard.document_loaders import KorLoader
from dreamsboard.dreams.aemo_representation_chain.prompts import (
    AEMO_REPRESENTATION_PROMPT_TEMPLATE,
) 

from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepNode
from typing import List
import logging
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class AEMORepresentationChain(ABC):
    aemo_representation_chain: Chain
    kor_dreams_task_step_chain: LLMChain
    def __init__(self,
                 start_task_context: str,
                 kor_dreams_task_step_chain: LLMChain,
                 aemo_representation_chain: Chain):
        self.start_task_context = start_task_context
        self.kor_dreams_task_step_chain = kor_dreams_task_step_chain
        self.aemo_representation_chain = aemo_representation_chain

    @classmethod
    def from_aemo_representation_chain(
            cls,
            llm_runable: Runnable[LanguageModelInput, BaseMessage],
            start_task_context: str,
            kor_dreams_task_step_llm: Runnable[LanguageModelInput, BaseMessage] = None,
    ) -> AEMORepresentationChain:
        # 00-判断情感表征是否符合.txt
        prompt_template1 = PromptTemplate(input_variables=["start_task_context"],
                                          template=os.environ.get(
                                "AEMO_REPRESENTATION_PROMPT_TEMPLATE", AEMO_REPRESENTATION_PROMPT_TEMPLATE
                            ))

        aemo_representation_chain = (prompt_template1 | llm_runable | StrOutputParser())
   
 
        def wrapper_output(_dict):
            return {
                # 中间变量全部打包输出
                "aemo_representation_context": _dict["aemo_representation_context"],
            }

        aemo_representation_chain = ({
                                        "aemo_representation_context": aemo_representation_chain,
                                    }
                                    | RunnableLambda(wrapper_output))


        kor_dreams_task_step_chain = KorLoader.form_kor_dreams_task_step_builder(
            llm_runable=llm_runable if kor_dreams_task_step_llm is None else kor_dreams_task_step_llm)
        return cls(start_task_context=start_task_context,
                   kor_dreams_task_step_chain=kor_dreams_task_step_chain,
                   aemo_representation_chain=aemo_representation_chain)

    def invoke_kor_dreams_task_step_context(self, aemo_representation_context: str) -> List[TaskStepNode]:
        """
        对开始任务进行抽取，得到任务步骤
        :return:
        """

        response = self.kor_dreams_task_step_chain.run(aemo_representation_context)
        task_step_node_list = []
        if response.get('data') is not None and response.get('data').get('script') is not None:
            step_list = response.get('data').get('script')
            for step in step_list:
                task_step_node = TaskStepNode(start_task_context=self.start_task_context,
                                            aemo_representation_context=aemo_representation_context,
                                            task_step_name=step.get('task_step_name'),
                                            task_step_description=step.get('task_step_description'),
                                            task_step_level=step.get('task_step_level'))
                task_step_node_list.append(task_step_node)

        return task_step_node_list
  
    def invoke_aemo_representation_context(self) -> Dict[str, Any]: 
       
        return self.aemo_representation_chain.invoke(
            {"start_task_context": self.start_task_context})
     


from dreamsboard.engine.constants import DATA_KEY, TYPE_KEY
from dreamsboard.engine.schema import (
    BaseNode,
)

from dreamsboard.engine.generate.code_generate import (
    CodeGenerator,
    BaseProgramGenerator,
    QueryProgramGenerator,
    AIProgramGenerator,
    EngineProgramGenerator,
)


def template_to_json(doc: BaseNode) -> dict:
    return {
        DATA_KEY: doc.dict(),
        TYPE_KEY: doc.get_type(),
    }


def json_to_template(template_dict: dict) -> BaseNode:
    template_type = template_dict[TYPE_KEY]
    data_dict = template_dict[DATA_KEY]
    doc: BaseNode

    if "extra_info" in data_dict:
        return legacy_json_to_template(template_dict)
    else:
        if template_type == BaseProgramGenerator.get_type():
            doc = BaseProgramGenerator.model_validate(data_dict)
        elif template_type == QueryProgramGenerator.get_type():
            doc = QueryProgramGenerator.model_validate(data_dict)
        elif template_type == AIProgramGenerator.get_type():
            doc = AIProgramGenerator.model_validate(data_dict)
        elif template_type == EngineProgramGenerator.get_type():
            doc = EngineProgramGenerator.model_validate(data_dict)
        else:
            raise ValueError(f"Unknown doc type: {template_type}")

        return doc


def legacy_json_to_template(template_dict: dict) -> CodeGenerator:
    """Todo: Deprecated legacy support for old node versions."""
    template_type = template_dict[TYPE_KEY]
    data_dict = template_dict[DATA_KEY]
    generator: CodeGenerator

    text = data_dict.get("text", "")
    _base_render_data = data_dict.get("render_data", {}) or {}
    id_ = data_dict.get("doc_id", None)

    if template_type == BaseProgramGenerator.get_type():
        generator = BaseProgramGenerator.from_config(cfg={
            "code_file": "base_template.py-tpl",
            "render_data": _base_render_data,
        })
    elif template_type == QueryProgramGenerator.get_type():
        generator = QueryProgramGenerator.from_config(cfg={
            "query_code_file": "dreams_query_template.py-tpl",
            "render_data": _base_render_data,
        })
    elif template_type == AIProgramGenerator.get_type():
        generator = AIProgramGenerator.from_config(cfg={
            "ai_code_file": "ai_template.py-tpl",
            "render_data": _base_render_data,
        })
    elif template_type == EngineProgramGenerator.get_type():
        generator = EngineProgramGenerator.from_config(cfg={
            "engine_code_file": "engine_template.py-tpl",
            "render_data": _base_render_data,
        })
    else:
        raise ValueError(f"Unknown template type: {template_type}")

    return generator



from typing import List, Optional

from langchain.schema.language_model import BaseLanguageModel

from langchain.chains import StuffDocumentsChain, LLMChain
from langchain_core.prompts  import PromptTemplate
from langchain.chains.combine_documents.map_reduce import ReduceDocumentsChain, MapReduceDocumentsChain

import sys
import asyncio
from langchain.docstore.document import Document


class DocumentWithVSId(Document):
    """
    矢量化后的文档
    """
    id: str = None


class SummaryAdapter:
    _OVERLAP_SIZE: int
    token_max: int
    _separator: str = "\n\n"
    chain: MapReduceDocumentsChain

    def __init__(self, overlap_size: int, token_max: int,
                 chain: MapReduceDocumentsChain):
        self._OVERLAP_SIZE = overlap_size
        self.chain = chain
        self.token_max = token_max

    @classmethod
    def form_summary(cls,
                     llm: BaseLanguageModel,
                     reduce_llm: BaseLanguageModel,
                     overlap_size: int,
                     token_max: int = 1300):
        """
        获取实例
        :param reduce_llm: 用于合并摘要的llm
        :param llm: 用于生成摘要的llm
        :param overlap_size: 重叠部分大小
        :param token_max: 最大的chunk数量，每个chunk长度小于token_max长度，第一次生成摘要时，大于token_max长度的摘要会报错
        :return:
        """

        # This controls how each document will be formatted. Specifically,
        document_prompt = PromptTemplate(
            input_variables=["page_content"],
            template="{page_content}"
        )

        # The prompt here should take as an input variable the
        # `document_variable_name`
        prompt_template = (
            "根据文本执行任务。以下任务信息"
            "{task_briefing}"
            "文本内容如下: "
            "\r\n"
            "{context}"
        )
        prompt = PromptTemplate(
            template=prompt_template,
            input_variables=["task_briefing", "context"]
        )
        llm_chain = LLMChain(llm=llm, prompt=prompt)
        # We now define how to combine these summaries
        reduce_prompt = PromptTemplate.from_template(
            "Combine these summaries: {context}"
        )
        reduce_llm_chain = LLMChain(llm=reduce_llm, prompt=reduce_prompt)

        document_variable_name = "context"
        combine_documents_chain = StuffDocumentsChain(
            llm_chain=reduce_llm_chain,
            document_prompt=document_prompt,
            document_variable_name=document_variable_name
        )
        reduce_documents_chain = ReduceDocumentsChain(
            token_max=token_max,
            combine_documents_chain=combine_documents_chain,
        )
        chain = MapReduceDocumentsChain(
            llm_chain=llm_chain,
            document_variable_name=document_variable_name,
            reduce_documents_chain=reduce_documents_chain,
            # 返回中间步骤
            return_intermediate_steps=True
        )
        return cls(overlap_size=overlap_size,
                   chain=chain,
                   token_max=token_max)

    def summarize(self,
                  file_description: str,
                  docs: List[DocumentWithVSId] = []
                  ) -> List[Document]:

        if sys.version_info < (3, 10):
            loop = asyncio.get_event_loop()
        else:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()

            asyncio.set_event_loop(loop)
        # 同步调用协程代码
        return loop.run_until_complete(self.asummarize(file_description=file_description,
                                                       docs=docs))

    async def asummarize(self,
                         file_description: str,
                         docs: List[DocumentWithVSId] = []) -> List[Document]:


        # TODO 暂不处理文档中涉及语义重复、上下文缺失、document was longer than the context length 的问题
        # merge_docs = self._drop_overlap(docs)
        # # 将merge_docs中的句子合并成一个文档
        # text = self._join_docs(merge_docs)
        # 根据段落于句子的分隔符，将文档分成chunk，每个chunk长度小于token_max长度

        """
        这个过程分成两个部分：
        1. 对每个文档进行处理，得到每个文档的摘要
         map_results = self.llm_chain.apply(
            # FYI - this is parallelized and so it is fast.
            [{self.document_variable_name: d.page_content, **kwargs} for d in docs],
            callbacks=callbacks,
        )
        2. 对每个文档的摘要进行合并，得到最终的摘要，return_intermediate_steps=True，返回中间步骤
        result, extra_return_dict = self.reduce_documents_chain.combine_docs(
            result_docs, token_max=token_max, callbacks=callbacks, **kwargs
        )
        """
        summary_combine, summary_intermediate_steps = self.chain.combine_docs(docs=docs,
                                                                              task_briefing="描述不同方法之间的接近度和相似性，"
                                                                                            "以帮助读者理解它们之间的关系。")

        # if len(summary_combine) == 0:
        #     # 为空重新生成，数量减半
        #     result_docs = [
        #         Document(page_content=question_result_key, metadata=docs[i].metadata)
        #         # This uses metadata from the docs, and the textual results from `results`
        #         for i, question_result_key in enumerate(
        #             summary_intermediate_steps["intermediate_steps"][
        #             :len(summary_intermediate_steps["intermediate_steps"]) // 2
        #             ])
        #     ]
        #     summary_combine, summary_intermediate_steps = self.chain.reduce_documents_chain.combine_docs(
        #         result_docs, token_max=self.token_max
        #     )

        doc_ids = ",".join([doc.id for doc in docs])
        _metadata = {
            "file_description": file_description,
            "summary_intermediate_steps": summary_intermediate_steps,
            "doc_ids": doc_ids
        }
        summary_combine_doc = Document(page_content=summary_combine, metadata=_metadata)

        return [summary_combine_doc]

    def _drop_overlap(self, docs: List[DocumentWithVSId]) -> List[str]:
        """
         # 将文档中page_content句子叠加的部分去掉
        :param docs:
        :param separator:
        :return:
        """
        merge_docs = []

        pre_doc = None
        for doc in docs:
            # 第一个文档直接添加
            if len(merge_docs) == 0:
                pre_doc = doc.page_content
                merge_docs.append(doc.page_content)
                continue

            # 列表中上一个结尾与下一个开头重叠的部分，删除下一个开头重叠的部分
            # 迭代递减pre_doc的长度，每次迭代删除前面的字符，
            # 查询重叠部分，直到pre_doc的长度小于 self._OVERLAP_SIZE // 2 - 2len(separator)
            for i in range(len(pre_doc), self._OVERLAP_SIZE // 2 - 2 * len(self._separator), -1):
                # 每次迭代删除前面的字符
                pre_doc = pre_doc[1:]
                if doc.page_content[:len(pre_doc)] == pre_doc:
                    # 删除下一个开头重叠的部分
                    merge_docs.append(doc.page_content[len(pre_doc):])
                    break

            pre_doc = doc.page_content

        return merge_docs

    def _join_docs(self, docs: List[str]) -> Optional[str]:
        text = self._separator.join(docs)
        text = text.strip()
        if text == "":
            return None
        else:
            return text


if __name__ == '__main__':

    docs = [

        '梦者有特别的作用，也就是说梦是在预卜未来。因此，梦内容的',

        '梦内容的多彩多姿以及对梦者本身所遗留的特殊印象，使他们很难想象',

        '使他们很难想象出一套系统划一的观念，而需要以其个别的价值与可靠性作各',
        '值与可靠性作各种不同的分化与聚合。因此，古代哲学家们对梦的评价也就完全'
    ]
    _OVERLAP_SIZE = 1
    separator: str = "\n\n"
    merge_docs = []
    # 将文档中page_content句子叠加的部分去掉，
    # 列表中上一个结尾与下一个开头重叠的部分，删除下一个开头重叠的部分
    pre_doc = None
    for doc in docs:
        # 第一个文档直接添加
        if len(merge_docs) == 0:
            pre_doc = doc
            merge_docs.append(doc)
            continue

        # 列表中上一个结尾与下一个开头重叠的部分，删除下一个开头重叠的部分
        # 迭代递减pre_doc的长度，每次迭代删除前面的字符，
        # 查询重叠部分，直到pre_doc的长度小于 _OVERLAP_SIZE-2len(separator)
        for i in range(len(pre_doc), _OVERLAP_SIZE - 2 * len(separator), -1):
            # 每次迭代删除前面的字符
            pre_doc = pre_doc[1:]
            if doc[:len(pre_doc)] == pre_doc:
                # 删除下一个开头重叠的部分
                page_content = doc[len(pre_doc):]
                merge_docs.append(page_content)

                pre_doc = doc
                break

    # 将merge_docs中的句子合并成一个文档
    text = separator.join(merge_docs)
    text = text.strip()

    print(text)



"""

"""

# 子任务指令转换为子问题.txt
CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE = """执行任务：结合开始任务（start_task_context），在符合任务总体描述（aemo_representation_context）的情况下，根据任务步骤名称（task_step_name）、任务步骤描述（task_step_description）和任务步骤层级（task_step_level），提出一个问题

### 任务

start_task_context: {start_task_context}

aemo_representation_context: {aemo_representation_context}


### 当前任务信息

task_step_name: {task_step_name}
task_step_description: {task_step_description}
task_step_level: {task_step_level}
"""


TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE = """改写问题为graphql语句，主要任务如下
1 、分析问题中概念，找出概念的语义词，
如 我想吃有煎蛋、火腿片和烤豆

concepts["food"] 

2、根据概念语义词(concepts) 分析，列出远离这个概念的任意一个词，靠近这个概念的若干词
如 我想吃有煎蛋、火腿片和烤豆

concepts["food"] 

moveAwayFrom: {{
  concepts: ["finance"], 
  force: 0.45
}}

moveTo: {{
  concepts: ["egg", "food"], 
  force: 0.85 
}}


3、 将概念和远离靠近的概念，生成"graphql语句" 任务

### 环境信息

RootQuery: "Get",
class: "{collection_name_context}", 
operators: "nearText" 

### Data Schema
```
(
    "{collection_name_context}",
     
    properties=[  # Define properties
        Property(name="refId", data_type=DataType.TEXT),
        Property(name="chunkText", data_type=DataType.TEXT), 
        
    ],
)
```

#### nearText Variables 

| Variable                 | Required | Type       | Description                                                  |
| ------------------------ | -------- | ---------- | ------------------------------------------------------------ |
| `concepts`               | yes      | `[string]` | An array of strings that can be natural language queries, or single words. If multiple strings are used, a centroid is calculated and used. Learn more about how the concepts are parsedhere. |
| `distance`               | no       | `float`    | The maximum allowed distance to the provided search input. Cannot be used together with the `certainty` variable. The interpretation of the value of the distance field depends on the distance metric used. |
| `certainty`              | no       | `float`    | Normalized Distance between the result item and the search vector. Normalized to be between 0 (perfect opposite) and 1 (identical vectors). Can't be used together with the `distance` variable. |
| `autocorrect`            | no       | `boolean`  | Autocorrect input text values. Requires the `text-spellcheck` module to be present & enabled. |
| `moveTo`                 | no       | `object<>` | Move your search term closer to another vector described by keywords |
| `moveTo<concepts>`       | no       | `[string]` | An array of strings - natural language queries or single words. If multiple strings are used, a centroid is calculated and used. |
| `moveTo<objects>`        | no       | `[UUID]`   | Object IDs to move the results to. This is used to "bias" NLP search results into a certain direction in vector space. |
| `moveTo<force>`          | no       | `float`    | The force to apply to a particular movement. Must be between 0 and 1 where 0 is equivalent to no movement and 1 is equivalent to largest movement possible. |
| `moveAwayFrom`           | no       | `object<>` | Move your search term away from another vector described by keywords |
| `moveAwayFrom<concepts>` | no       | `[string]` | An array of strings - natural language queries or single words. If multiple strings are used, a centroid is calculated and used. |
| `moveAwayFrom<objects>`  | no       | `[UUID]`   | Object IDs to move the results from. This is used to "bias" NLP search results into a certain direction in vector space. |
| `moveAwayFrom<force>`    | no       | `float`    | The force to apply to a particular movement. Must be between 0 and 1 where 0 is equivalent to no movement and 1 is equivalent to largest movement possible. |


### Semantic Path

Example: showing a semantic path without edges.
```

_additional {{
  semanticPath {{
    path {{
      concept
      distanceToNext
      distanceToPrevious
      distanceToQuery
      distanceToResult
    }}
  }}
}}

```


### 注意
请不要修改环境信息


### 结果示例
```
{{
  Get {{
    {collection_name_context}(
      nearText:{{
        concepts: ["food"], 
        distance: 0.23, 
        moveAwayFrom: {{
          concepts: ["finance"],
          force: 0.45
        }},
        moveTo: {{
          concepts: ["egg", "food"],
          force: 0.85
        }}
      }}, 
      limit: 25
    ) {{
     
     refId
     chunkText
     _additional {{
        semanticPath {{
          path {{
            concept
            distanceToNext
            distanceToPrevious
            distanceToQuery
            distanceToResult
          }}
        }}
      }}
    }}
  }}
}}

```



下面执行这个问题相关的任务
{task_step_question}
"""
 


 




import os
import fsspec
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from dataclasses_json import DataClassJsonMixin
from typing import Any, Dict, List, Optional, Sequence

from dreamsboard.engine.entity.task_step.task_step import TaskStepNode

DEFAULT_PERSIST_FNAME = "task_step_store.json"
DEFAULT_PERSIST_DIR = "./storage"
DEFAULT_PERSIST_PATH = os.path.join(DEFAULT_PERSIST_DIR, DEFAULT_PERSIST_FNAME)


@dataclass
class RefTaskStepInfo(DataClassJsonMixin):
    """Dataclass to represent ingested task steps."""

    node_ids: List = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)


class BaseTaskStepStore(ABC):
    # ===== ，加载与保存抽象定义Save/load =====
    def persist(
        self,
        persist_path: str = DEFAULT_PERSIST_PATH,
        fs: Optional[fsspec.AbstractFileSystem] = None,
    ) -> None:
        """Persist the docstore to a file."""
        pass

    # ===== 存储任务步骤相关的信息 =====
    @property
    @abstractmethod
    def task_step_all(self) -> Dict[str, TaskStepNode]:
        ...

    @abstractmethod
    def add_task_step(
        self, task_steps: Sequence[TaskStepNode], allow_update: bool = True
    ) -> None:
        ...

    @abstractmethod
    def get_task_step(self, task_step_id: str, raise_error: bool = True) -> Optional[TaskStepNode]:
        ...

    @abstractmethod
    def delete_task_step(self, task_step_id: str, raise_error: bool = True) -> None:
        """Delete a task_step from the store."""
        ...

    @abstractmethod
    def task_step_exists(self, task_step_id: str) -> bool:
        ...

    # ==== 存储任务步骤相关的信息引用节点 =====
    @abstractmethod
    def get_all_ref_task_step_info(self) -> Optional[Dict[str, RefTaskStepInfo]]:
        """Get a mapping of ref_task_step_id -> RefTaskStepInfo for all ingested nodes."""

    @abstractmethod
    def get_ref_task_step_info(self, ref_task_step_id: str) -> Optional[RefTaskStepInfo]:
        """Get the RefTaskStepInfo for a given ref_task_step_id."""

    @abstractmethod
    def delete_ref_task_step(self, ref_task_step_id: str, raise_error: bool = True) -> None:
        """Delete a ref_task_step and all it's associated nodes."""

    @abstractmethod
    def ref_task_step_exists(self, ref_task_step_id: str) -> bool:
        """Check if a ref_task_step_id has been ingested."""




COSPLAY_ANALYSIS_MD_TEMPLATE = """


### 00_个人ip介绍
> 昵称：{cosplay_role}
>
> 资源文件({source_url})
>
> 字幕文件([{keyframe}]({keyframe_path}))
>
> 存储文件([{storage_keyframe}]({storage_keyframe_path}))


#### 03- 故事情境生成 `story_scenario_context`
```text
{story_scenario_context}
```

#### 03-故事场景生成 `scene_monologue_context`
```text
{scene_monologue_context}
```

#### 04-情感情景引导
```text
{dreams_gen_text}
```

#### 04-情感情景引导-开放性问题 `dreams_guidance_context`
```text
{dreams_guidance_context}
```


#### 05-剧情总结 `evolutionary_step`
```text
{evolutionary_step}
```

#### 05-性格分析 `dreams_personality_context`
```text
{dreams_personality_context}
```

"""



from typing import Any, List, Optional, Sequence

from dreamsboard.engine.engine_builder import BaseEngineBuilder
from dreamsboard.engine.registry import INDEX_STRUCT_TYPE_TO_INDEX_CLASS
from dreamsboard.engine.storage.storage_context import StorageContext

import logging

logger = logging.getLogger(__name__)


def load_store_from_storage(
        storage_context: StorageContext,
        index_id: Optional[str] = None,
        **kwargs: Any,
) -> BaseEngineBuilder:
    """Load index from storage context.

    Args:
        :param index_id: index_id(Optional[Sequence[str]]): IDs of the indices to load.
            Defaults to None, which loads all indices in the index store.
        :param storage_context: storage_context (StorageContext): storage context containing
            template store, index store

        **kwargs: Additional keyword args to pass to the index constructors.
    """
    index_ids: Optional[Sequence[str]]
    if index_id is None:
        index_ids = None
    else:
        index_ids = [index_id]

    indices = load_indices_from_storage(storage_context, index_ids=index_ids, **kwargs)

    if len(indices) == 0:
        raise ValueError(
            "No index in storage context, check if you specified the right persist_dir."
        )
    elif len(indices) > 1:
        raise ValueError(
            f"Expected to load a single index, but got {len(indices)} instead. "
            "Please specify index_id."
        )

    return indices[0]


def load_indices_from_storage(
        storage_context: StorageContext,
        index_ids: Optional[Sequence[str]] = None,
        **kwargs: Any,
) -> List[BaseEngineBuilder]:
    """Load multiple indices from storage context

    Args:
        :param storage_context: storage_context (StorageContext): storage context containing
            template store, index store
        :param index_ids: index_id(Optional[Sequence[str]]): IDs of the indices to load.
            Defaults to None, which loads all indices in the index store.

        **kwargs: Additional keyword args to pass to the index constructors.
    """
    if index_ids is None:
        logger.info("Loading all indices.")
        index_structs = storage_context.index_store.index_structs()
    else:
        logger.info(f"Loading indices with ids: {index_ids}")
        index_structs = []
        for index_id in index_ids:
            index_struct = storage_context.index_store.get_index_struct(index_id)
            if index_struct is None:
                raise ValueError(f"Failed to load index with ID {index_id}")
            index_structs.append(index_struct)

    indices = []
    for index_struct in index_structs:
        type_ = index_struct.get_type()
        index_cls = INDEX_STRUCT_TYPE_TO_INDEX_CLASS[type_]
        index = index_cls(
            index_struct=index_struct, storage_context=storage_context, **kwargs
        )
        indices.append(index)
    return indices



import os

from llama_index.llms.openai import OpenAI
from llama_index.embeddings.openai import (
    OpenAIEmbedding,
)  # pants: no-infer-dep

from dreamsboard.agent.react_agent_build_query_engine_tools import build_query_engine_tools, build_query_docs, \
    build_query_docs_index_store
from dreamsboard.agent.react_agent_finetune_cleaning import adapter_agent, build_dataset_generator_questions, \
    gen_question_variations
from dreamsboard.agent.react_tools_utils import save_questions

if __name__ == '__main__':
    # 获取当前文件路径
    root_dir = os.path.dirname(os.path.abspath(__file__))
    llm = OpenAI(model="gpt-4", temperature=0.3, api_key="sk-ApUK41y73g8qMbrz36A81641752946449f10BbBe32Ff2b7c",
                 api_base="http://localhost:3000/v1")
    embeddings = OpenAIEmbedding(api_key="EMPTY", api_base="http://127.0.0.1:9997/v1")
    query_docs = build_query_docs()
    index_store = build_query_docs_index_store(query_docs, embeddings)
    query_engine_tools = build_query_engine_tools(llm, index_store)
    agent_llm = OpenAI(model="gpt-4", temperature=0.9, api_key="sk-ApUK41y73g8qMbrz36A81641752946449f10BbBe32Ff2b7c",
                       api_base="http://localhost:3000/v1")
    base_agent = adapter_agent(agent_llm, query_engine_tools)

    response = base_agent.chat(
        "查询组织机构的接口是什么"
    )
    print(str(response))
    base_question_gen_query = (
        "你是一名公司业务产品，你的任务是基于系统设计访问的思想设计一套电网拓扑、图形、资源、资产、测点等基础业务。"
        "使用业务接口文档提供的上下文， 制定一些问题，"
        "从上下文中捕捉到重要事实形成问题，"
        "将问题限制在所提供的上下文信息内."
        "**提取的事实需要验证,不需要标注出具体的来源和分类标签**"
        "例如："
        "  停/供电范围分析，在拓扑服务吗？"
        "  问题"
        "**请注意你只需要查看接口文档中的业务介绍、请求路径、请求方式，请忽略请求参数和示例信息、返回参数、调用范例**"
    )
    num = 360
    # 拼接 所有query_docs
    all_docs = []
    for doc in query_docs:
        all_docs.extend(doc)

    questions = build_dataset_generator_questions(base_question_gen_query=base_question_gen_query,
                                                  docs=all_docs,
                                                  llm=llm, num=num)
    print(len(questions))
    print(questions)
    save_questions(questions, os.path.join(root_dir, f"glm_questions_{num}q.txt"))
    vary_question_tmpl = """\
你是一位公司产品经理。给定一个关于接口文档的产品需求的问题，你的目标是生成多达 {num_vary} 个问题变体，涉及多个接口文档 。

这可能包括比较/对比不同接口文档，你可以通过业务介绍、请求路径、请求方式生成，或只能通过两个业务介绍的问题（发挥创意！）

你被提供了一组有效的接口文档。请仅生成可以在该组接口文档中回答的问题变体。

For example:
Base Question: 如何通过《OOS文件系统业务介绍》上传文件？
Valid 10Qs: [《组织人员业务接口》, 《任务调度系统接口》, 《OOS文件系统业务介绍》]
Question Variations: 
使用《任务调度系统接口》创建任务后，如何利用《OOS文件系统业务介绍》对任务结果进行存储？
在《组织人员业务接口》中添加新员工后，如何配置《OOS文件系统业务介绍》以分配个人文件存储空间？  
如何结合使用《OOS文件系统业务介绍》和《任务调度系统接口》来优化文件的自动备份流程？  

现在让我们试试吧！

Base Question: {base_question}
Valid 10Qs: {valid_10qs}
Question Variations:
"""

    VALID_10Q_STR = "[关于测点管理中心的业务中台, 关于电网拓扑中心的业务中台, 关于电网图形中心的业务中台, 关于电网资产中心的业务中台, 关于电网资源中心的业务中台, 关于电基础服务中心的业务中台]"

    new_questions = gen_question_variations(vary_question_tmpl=vary_question_tmpl, valid_10q_str=VALID_10Q_STR,
                                            generator_questions=questions, llm=llm, num_vary=3)
    print(len(new_questions))
    print(new_questions)
    total_questions = num * (3 + 1)
    # 训练集和验证集，训练集占80%，验证集占20%
    train_questions, eval_questions = new_questions[:int(total_questions * 0.8)], new_questions[
                                                                                  int(total_questions * 0.8):]

    save_questions(train_questions, os.path.join(root_dir, f"glm_train_questions_{int(total_questions * 0.8)}q.txt"))
    save_questions(eval_questions, os.path.join(root_dir, f"glm_eval_questions_{int(total_questions * 0.2)}q.txt"))



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Any, Optional, Dict, List

import logging

from pydantic import Field

from dreamsboard.engine.schema import BaseNode, ObjectTemplateType
from dreamsboard.templates import get_template_path
from pydantic import BaseModel
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

class TaskStepContext(BaseModel):
    ref_id: str = Field(default="", description="ref_id")
    score: float = Field(default=0.0, description="score")
    text: str = Field(default="", description="text")

class TaskStepNode(BaseNode, ABC):
    """任务步骤节点"""
    start_task_context: Optional[str] = Field(
        default="", description="开始任务"
    ) 
    aemo_representation_context: Optional[str] = Field(
        default="", description="任务总体描述"
    )
    task_step_name: Optional[str] = Field(
        default="", description="任务步骤名称"
    ) 
    task_step_description: Optional[str] = Field(
        default="", description="任务步骤描述"
    )
    task_step_level: Optional[str] = Field(
        default="", description="任务步骤层级"
    )
    task_step_question: Optional[str] = Field(
        default="", description="任务步骤问题"
    )
    task_step_question_context: Optional[List[TaskStepContext]] = Field(
        default=[], description="任务步骤问题上下文"
    )

    task_step_question_answer: Optional[str] = Field(
        default="", description="任务步骤答案"
    )
    ref_task_step_id: Optional[str] = Field(
        default="", description="ref_task_step_id",
    )

    def __init__(self,
                 start_task_context: str = None,
                 aemo_representation_context: str = None,
                 task_step_name: str = None,
                 task_step_description: str = None,
                 task_step_level: str = None,
                 task_step_question: str = None,
                 task_step_question_context: List[TaskStepContext] = None,
                 task_step_question_answer: str = None,
                 ref_task_step_id: str = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.start_task_context = start_task_context
        self.aemo_representation_context = aemo_representation_context
        self.task_step_name = task_step_name
        self.task_step_description = task_step_description
        self.task_step_level = task_step_level
        self.task_step_question = task_step_question
        self.task_step_question_context = task_step_question_context
        self.task_step_question_answer = task_step_question_answer
        self.ref_task_step_id = ref_task_step_id
    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        start_task_context = cfg.get("start_task_context", "") 
        aemo_representation_context = cfg.get("aemo_representation_context", "")
        task_step_name = cfg.get("task_step_name", "")
        task_step_description = cfg.get("task_step_description", "")
        task_step_level = cfg.get("task_step_level", "")
        task_step_question = cfg.get("task_step_question", "")
        task_step_question_context = cfg.get("task_step_question_context", [])
        task_step_question_answer = cfg.get("task_step_question_answer", "")
        ref_task_step_id = cfg.get("ref_task_step_id", "")
        return cls(
            start_task_context=start_task_context,
            aemo_representation_context=aemo_representation_context,
            task_step_name=task_step_name,
            task_step_description=task_step_description,
            task_step_level=task_step_level,
            task_step_question=task_step_question,
            task_step_question_context=task_step_question_context,
            task_step_question_answer=task_step_question_answer,
            ref_task_step_id=ref_task_step_id
        )

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return "task_step_node"

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            return "任务步骤节点"

    @property
    def template_content(self) -> str:
        return f"start_task_context: {self.start_task_context}, " \
               f"aemo_representation_context: {self.aemo_representation_context}, " \
               f"task_step_name: {self.task_step_name}, " \
               f"task_step_description: {self.task_step_description}, " \
               f"task_step_level: {self.task_step_level}, " \
               f"task_step_question: {self.task_step_question}, " \
               f"task_step_question_context: {self.task_step_question_context}, " \
               f"task_step_question_answer: {self.task_step_question_answer}, " \
               f"ref_task_step_id: {self.ref_task_step_id}"

    @template_content.setter
    def template_content(self, _template_content) -> None:
        raise RuntimeError("template_content is read only.")



import logging

import langchain

from dreamsboard.dreams.task_step_md.prompts import TASK_MD_TEMPLATE, TASK_STEP_MD_TEMPLATE
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from langchain_core.runnables import RunnableParallel, RunnableLambda
from langchain_core.prompt_values import StringPromptValue

from langchain.prompts import PromptTemplate

logger = logging.getLogger(__name__)

_PROMPT_TEMPLATE_1 = PromptTemplate(input_variables=[
    "task_step_name",
    "task_step_level",
    "task_step_id",
    "task_step_question_answer"],
    template=TASK_STEP_MD_TEMPLATE)
_PROMPT_TEMPLATE_2 = PromptTemplate(
    input_variables=[
        "start_task_context",
        "aemo_representation_context",
        "context_placeholder"
    ],
    template=TASK_MD_TEMPLATE)


class TaskStepMD:

    def __init__(self,
                 task_step_store: BaseTaskStepStore
                 ):
        self.task_step_store = task_step_store

    def format_md(self) -> StringPromptValue:
 
        
        def wrapper_steps_unit(dict_input: dict):
            # 使用 TASK_STEP_MD_TEMPLATE 格式化每个任务步骤
            formatted_task_steps = [
            ]
            for step in list(self.task_step_store.task_step_all.values()):
                
                if '>' in step.task_step_level: 
                    
                    step_text = _PROMPT_TEMPLATE_1.format(
                        task_step_name=step.task_step_name,
                        task_step_level=step.task_step_level,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer
                    )
                    formatted_task_steps.append('   '+step_text.strip())
                else:
                    step_text = _PROMPT_TEMPLATE_1.format(
                        task_step_name=f'**{step.task_step_name}**',
                        task_step_level=step.task_step_level,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer
                    )
                    formatted_task_steps.append('\n' + step_text.strip())

            # 将格式化的步骤列表转换为字符串
            context_placeholder = "".join(formatted_task_steps)
            return {
                "start_task_context": list(self.task_step_store.task_step_all.values())[0].start_task_context,
                "aemo_representation_context": list(self.task_step_store.task_step_all.values())[0].aemo_representation_context,
                "context_placeholder": context_placeholder
            }
 

        chain = (RunnableLambda(wrapper_steps_unit)  
                 | _PROMPT_TEMPLATE_2)

        out = chain.invoke({})
        return out

    def write_md(self, output_path: str) -> StringPromptValue:
        md = self.format_md()
        with open(output_path, "w") as f:
            f.write(md.text)
        logger.info(f"Write MD to {output_path}")
        return md



from pydantic import BaseModel, Field


class PromptConfig(BaseModel):
    critic_system_prompt: str
    critic_system_prompt_data: str
    refine_system_prompt: str
    refine_system_prompt_data: str
    evaluate_system_prompt: str
    evaluate_system_prompt_data: str
 

class RefineResponse(BaseModel):
    thought: str = Field(..., description="The thought process behind the answer.")
    answer: str = Field(..., description="The thought process behind the answer.")
    answer_score: float = Field(..., description="The answer to the problem.")


gpt_prompt_config = PromptConfig(
    critic_system_prompt="""完成你的目标任务,输出详细且有建设性的批评意见以改进`<current_answer>`， step by step plan. 

# 补充指南  

- 不要重复`<problem>`描述。  
- 不要重复`<current_answer>`描述。
- 不要重复`<start_task_context>`描述。
- 不要重复`<aemo_representation_context>`描述。
- 不要重复`<task_step_name>`描述。
- 不要重复`<task_step_description>`描述。
- 不要重复`<task_step_level>`描述。


结合开始任务（start_task_context），在符合任务总体描述（aemo_representation_context）的情况下，根据任务步骤名称（task_step_name）、任务步骤描述（task_step_description）和任务步骤层级（task_step_level），

突出显示需要改进或更正的特定区域。不需要更多步骤, 不要将之前完成的步骤作为计划的一部分返回。
""",
    critic_system_prompt_data="""你的目标:
<problem>
{problem}
</problem>

你目前的结果在这里:
<context>
{context}
</context>

<current_answer>
{current_answer}
</current_answer>

你目前已完成以下步骤：
{past_steps}


# 参考资源

start_task_context: {start_task_context}
aemo_representation_context: {aemo_representation_context}

# 当前任务信息

task_step_name: {task_step_name}
task_step_description: {task_step_description}
task_step_level: {task_step_level}
""",
    refine_system_prompt="""# 目标
根据批评意见 (`<critique>`) 优化当前回答 (`<current_answer>`) 并续写上下文 （`<context>`）。

### 补充指南  

- 不要重复`<problem>`描述。  
- 不要重复`<current_answer>`描述。 
- 结合总体任务进度输出 
- 不要做总结的回复

# 输出优化后的回答

### 优化后的回答: 输出的优化答案 
输出要求 
    - 文本不超过500字

### 续写的上下文: 包含续写的上下文
输出要求
    - 不允许出现`标题`、`分段`、`综上所述`的文字

### 评分分数：包含续写的评分分数
输出要求
    - 格式`分数/总分`

""",
refine_system_prompt_data="""# 当前上下文：
<context>
{context}
</context>

# 当前回答：
<current_answer>
{current_answer}
</current_answer>

# 问题描述：
<problem>
{problem}
</problem>

# 批评意见：
<critique>
{critique}
</critique>

# 已完成的步骤：
{past_steps}
""",
    evaluate_system_prompt="""# Provide a reward score between -100 and 100 for the answer quality, using very strict standards. 
- Do not give a full score above 95. Ensure the reward score is an integer.
- Does not want answers with numbered or structured content like 1. 2. 3. 4. 
- Avoid using any segmented or explicitly structured terms. If disallowed content appears, assign a score of 10. 
- Return *ONLY* the score. 

""",
    evaluate_system_prompt_data="""<problem>
{problem}
</problem>
<answer>
{answer}
</answer>
""",
)



# 任务引擎构建器核心类，用于构建任务引擎
 
# 3、对每个子任务载入会话场景，然后按照扩写任务步骤构建，MCTS任务

from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from dreamsboard.engine.engine_builder import CodeGeneratorBuilder

from dreamsboard.dreams.task_step_to_question_chain.base import TaskStepToQuestionChain
from dreamsboard.dreams.task_step_to_question_chain.weaviate.context_collections import init_context_connect

from dreamsboard.document_loaders import StructuredStoryboardCSVBuilder
from dreamsboard.dreams.builder_cosplay_code.base import StructuredDreamsStoryboard
from dreamsboard.dreams.dreams_personality_chain.base import StoryBoardDreamsGenerationChain
from dreamsboard.document_loaders.structured_storyboard_loader import StructuredStoryboard

from dreamsboard.engine.entity.dreams_personality.dreams_personality import DreamsPersonalityNode
from dreamsboard.engine.generate.code_generate import QueryProgramGenerator, EngineProgramGenerator
from dreamsboard.engine.loading import load_store_from_storage
from dreamsboard.engine.storage.dreams_analysis_store.simple_dreams_analysis_store import SimpleDreamsAnalysisStore
from dreamsboard.engine.storage.storage_context import StorageContext
from dreamsboard.engine.utils import concat_dirs
import logging 
from dreamsboard.engine.memory.mctsr.mctsr import MCTSNode, MCTSrStoryboard
from langchain.schema import AIMessage
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class TaskEngineBuilder:

    """TaskEngineBuilder 场景加载模块
		执行会话场景资源初始化，构建MCTS任务

    根据任务步骤，构建场景加载模块，生成资源文件csv
    根据每个任务，载入StructuredDreamsStoryboard 会话场景
    按照扩写任务步骤构建MCTS任务
    
		输入：
			task_step_id
			task_step_store: 任务存储器（SimpleTaskStepStore）
			start_task_context： 起始任务
			llm： 模型

    
    """
    base_path: str
    cross_encoder_path: str
    storage_context: StorageContext
    task_step_store: BaseTaskStepStore
    task_step_id: str
    csv_file_path: str
    storyboard_executor: StructuredDreamsStoryboard
    llm_runable: Runnable[LanguageModelInput, BaseMessage]

    def __init__(self,
                base_path: str,
                llm_runable: Runnable[LanguageModelInput, BaseMessage],
                cross_encoder_path: str,
                start_task_context: str,
                task_step_store: BaseTaskStepStore,
                task_step_id: str,
    ):
        self.base_path = base_path
        self.start_task_context = start_task_context
        self.task_step_store = task_step_store
        self.task_step_id = task_step_id
        self.llm_runable = llm_runable
        self.cross_encoder_path = cross_encoder_path
        self.client = None
        self.task_step_to_question_chain = None
        self.csv_file_path = None
        self.storyboard_executor = None
        self.storage_context = StorageContext.from_defaults(
            persist_dir=f"{self.base_path}/storage/{self.task_step_id}"
        )
        self.storage_context.task_step_store = self.task_step_store

    def check_engine_init(self):
        """
        检查引擎是否初始化
        """
        try:
            storage_context = StorageContext.from_defaults(
                persist_dir=f"{self.base_path}/storage/{self.task_step_id}"
            )
            if storage_context.task_step_store is None:
                return False
            if storage_context.dreams_analysis_store is None:
                return False
            if storage_context.template_store is None:
                return False
            if storage_context.index_store is None:
                return False
            
            code_gen_builder = load_store_from_storage(storage_context=storage_context)
            if code_gen_builder is None:
                return False
            else:
                return True
        except:
            return False

    def init_task_engine(self):
        """
        
        初始化任务引擎
        》TaskStepToQuestionChain 
        	输入：
        		client： 矢量库客户端
        		llm： 模型
					invoke_task_step_to_question：1、 对开始任务进行抽取，得到任务步骤，提示词所要求的输入拆分成子任务， 
					invoke_task_step_question_context： 2、对每个子任务指令转换为子问题，召回问题前3条，对任务步骤进行抽取，得到任务步骤的上下文
					export_csv_file_path: 3、对召回内容与问题 导出csv文件
 
        """
              
        client = init_context_connect()


        self.task_step_to_question_chain = TaskStepToQuestionChain.from_task_step_to_question_chain(
            base_path=self.base_path,
            start_task_context=self.start_task_context,
            llm_runable=self.llm_runable, 
            task_step_store=self.task_step_store,
            client=client,
            cross_encoder_path=self.cross_encoder_path
        )
        self.task_step_to_question_chain.invoke_task_step_to_question(self.task_step_id)
        self.task_step_to_question_chain.invoke_task_step_question_context(self.task_step_id)
        self.csv_file_path = self.task_step_to_question_chain.export_csv_file_path(self.task_step_id)
   
 
    def init_task_engine_dreams(self, allow_init: bool = True) -> None:
        """
        初始化场景加载资源StoryBoardDreamsGenerationChain
        如果allow_init为True，则清空存储，重新初始化
     
        对每个子任务通过职业提示词，载入会话场景
            1、构建场景信息（story_scenario_context），提示词（STORY_BOARD_SCENE_TEMPLATE）
            2、对任务上下文(story_board_summary_context)，构建第一人称数据(scene_monologue_context),提示词（STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE）
            3、对任务上下文(story_board_summary_context)，获取任务分析(evolutionary_step), 提示词（EDREAMS_EVOLUTIONARY_TEMPLATE）
            4、对任务分析(evolutionary_step)，分析对话预设信息（性格）， 提示词（EDREAMS_PERSONALITY_TEMPLATE）
            5、对任务上下文(story_board_summary_context)，场景信息story_scenario_context, 第一人称数据(scene_monologue_context)，
            生成关于人物职业的引导话术，提示词（DREAMS_GEN_TEMPLATE）
        """ 
        if self.csv_file_path is None:
            raise ValueError("csv_file_path is None, please invoke init_task_engine first")
        
        dreams_analysis_store = SimpleDreamsAnalysisStore.from_persist_dir(persist_dir=f"{self.base_path}/storage/{self.task_step_id}")
        if allow_init:
            analysis_ids = list(dreams_analysis_store.analysis_all.keys())
            for analysis_id in analysis_ids:
                dreams_analysis_store.delete_analysis(analysis_id=analysis_id)

            dreams_generation_chain = StoryBoardDreamsGenerationChain.from_dreams_personality_chain(
                llm_runable=self.llm_runable, csv_file_path=self.csv_file_path
            )

            output = dreams_generation_chain.run() 
            # 拼接dreams_guidance_context和dreams_personality_context两个字典
            dreams_generation = {}
            dreams_generation.update(output.get("dreams_guidance_context"))
            dreams_generation.update(output.get("dreams_personality_context"))
            dreams = DreamsPersonalityNode.from_config(cfg=dreams_generation)
            dreams_analysis_store.add_analysis([dreams])
            dreams_analysis_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{self.task_step_id}", basename="dreams_analysis_store.json")
            dreams_analysis_store.persist(persist_path=dreams_analysis_store_path)

        self.storage_context.dreams_analysis_store = dreams_analysis_store
        self.storage_context.persist(persist_dir=f"{self.base_path}/storage/{self.task_step_id}")
            
    def init_task_engine_storyboard_executor(self) -> None:
        """
        构建会话场景执行器
        """
        
        if self.csv_file_path is None:
            raise ValueError("csv_file_path is None, please invoke init_task_engine first")
        
        if self.storage_context is None or self.storage_context.dreams_analysis_store is None:
            raise ValueError("storage_context is None or dreams_analysis_store is None, please invoke init_task_engine_dreams first")
        
        for val in self.storage_context.dreams_analysis_store.analysis_all.values():
            dreams_guidance_context = val.dreams_guidance_context
            dreams_personality_context = val.dreams_personality_context

        csv_builder = StructuredStoryboardCSVBuilder.form_builder(csv_file_path=self.csv_file_path)
        csv_builder.load()
        self.storyboard_executor = StructuredDreamsStoryboard.form_builder(
            llm_runable=self.llm_runable,
            builder=csv_builder,
            dreams_guidance_context=dreams_guidance_context,
            dreams_personality_context=dreams_personality_context,
            guidance_llm=self.llm_runable,
            personality_llm=self.llm_runable,
            user_id=self.task_step_id
        )

    def storyboard_code_gen_builder(self) -> CodeGeneratorBuilder:
        """
        构建会话场景执行器
        """
        if self.storage_context is None:
            raise ValueError("storage_context is None, please invoke init_task_engine first")
        
        try:
            code_gen_builder = load_store_from_storage(storage_context=self.storage_context)
        except:
            code_gen_builder = self.storyboard_executor.loader_cosplay_builder(
                dreams_cosplay_role=self.task_step_id,
                storage_context=self.storage_context, 
            )
            
        code_gen_builder.storage_context.dreams_analysis_store = self.storage_context.dreams_analysis_store
        code_gen_builder.storage_context.template_store = self.storage_context.template_store
        code_gen_builder.storage_context.index_store = self.storage_context.index_store
        code_gen_builder.storage_context.task_step_store = self.storage_context.task_step_store
        code_gen_builder.storage_context.persist(persist_dir=f"{self.base_path}/storage/{self.task_step_id}")

        return code_gen_builder
    
    def generate_step_answer(self, code_gen_builder: CodeGeneratorBuilder) -> str:
        """
        生成当前任务的答案
        """
        task_step = self.task_step_store.get_task_step(self.task_step_id)
        ai_message = self._get_ai_message(
            user_prompt=task_step.task_step_question,
            code_gen_builder=code_gen_builder, 
        )
        logger.info(ai_message.content)
        task_step.task_step_question_answer = ai_message.content
        self.task_step_store.add_task_step([task_step])
        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{self.task_step_id}", basename=DEFAULT_PERSIST_FNAME)
        self.task_step_store.persist(persist_path=task_step_store_path) 
        return ai_message.content
    
    def _get_ai_message(self, 
                        user_prompt: str,
                        code_gen_builder: CodeGeneratorBuilder,
        ) -> AIMessage:
        """
        获取AI消息
        """
                 
        code_gen_builder.add_generator(QueryProgramGenerator.from_config(cfg={
            "query_code_file": "query_template.py-tpl",
            "render_data": {
                'cosplay_role': 'user',
                'message': user_prompt
            },
        })) 
        executor = code_gen_builder.build_executor(
            llm_runable=self.llm_runable,
            messages=[]
        )
        logger.info(executor.executor_code)

        executor.execute()
        _ai_message = executor.chat_run()

        logger.info(executor._ai_message)
        assert executor._ai_message is not None
        
        code_gen_builder.remove_last_generator()

        return _ai_message


    def get_mcts_node(self) -> MCTSrStoryboard:
        """
        构建MCTS树, 初始化当前任务相关的MCTS节点，并返回MCTS执行器
        """
        task_step_all = self.task_step_store.task_step_all
        task_step_all_list = [val.__dict__ for val in list(task_step_all.values())]
        structured_storyboard = StructuredStoryboard(json_data=task_step_all_list) 
        linked_list_node = structured_storyboard.get_task_step_node(self.task_step_id)
        
        mcts_node = MCTSNode(
            base_path=self.base_path,
            answer=linked_list_node.task_step_question_answer,
            linked_list_node=linked_list_node,
            storage_context=self.storage_context, 
            parent=None, 
            children=[], 
            visits=0, 
            Q=0, 
            reward_samples=[]
        )
        # 构建MCTS树, 初始化当前节点的顶层节点
        while linked_list_node is not None and linked_list_node.prev is not None:
            linked_list_node = linked_list_node.prev
            parent_node = MCTSNode(
                base_path=self.base_path,
                answer=linked_list_node.task_step_question_answer, 
                linked_list_node=linked_list_node,
                storage_context=self.storage_context, 
                parent=None, 
                children=[], 
                visits=0, 
                Q=0, 
                reward_samples=[]
            )
            mcts_node.parent = parent_node

            
        task_step = self.task_step_store.get_task_step(self.task_step_id)
        mctsr = MCTSrStoryboard.model_construct(
            llm_runable=self.llm_runable,
            problem=task_step.task_step_name, 
            max_rollouts=2
        )
        mctsr.initialize(mcts_node)
        return mctsr



from __future__ import annotations
from abc import ABC
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from langchain.chains.base import Chain
from langchain_core.prompts  import PromptTemplate
from langchain.chains import SequentialChain
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableParallel, RunnableLambda

from dreamsboard.document_loaders import KorLoader

from dreamsboard.engine.storage.storage_context import BaseTaskStepStore

from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepNode
from dreamsboard.dreams.task_step_to_question_chain.prompts import (
    CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE,
    TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE
)

from dreamsboard.engine.entity.task_step.task_step import TaskStepContext

from dreamsboard.dreams.task_step_to_question_chain.weaviate.context_collections import init_context_collections
from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import exe_query, get_query_hash
from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import insert_into_database
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import find_root_nodes
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import find_high_outdegree_concepts
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import find_high_indegree_nodes
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import create_interactive_graph
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import find_simple_path
from dreamsboard.dreams.task_step_to_question_chain.weaviate.init_networkx_concept import create_G

import json
import copy
from sentence_transformers import CrossEncoder
from weaviate.client import WeaviateClient
from weaviate.classes.query import Filter, Rerank, MetadataQuery
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME
from dreamsboard.engine.utils import concat_dirs
from typing import List
import logging
import os
import re
import pandas as pd
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

"""
#### 场景加载模块

编写符合计算机科学领域的 故事情境提示词，生成研究情境（story_scenario_context），替换现有的langchain会话模板，
对每个子任务指令转换为子问题
召回问题前3条,存入task_step_question_context
调用llm，生成task_step_question_answer
"""

PATTERN = re.compile(r"```graphql?([\s\S]*?)```", re.DOTALL)
"""Regex pattern to parse the output."""

class TaskStepToQuestionChain(ABC):
    base_path: str
    start_task_context: str
    collection_name: str
    collection_name_context: str
    task_step_to_question_chain: Chain
    task_step_store: BaseTaskStepStore
    task_step_question_to_graphql_chain: Chain
    client: WeaviateClient
    cross_encoder: CrossEncoder
    def __init__(self,
                 base_path: str,
                 start_task_context: str,
                 collection_name: str,
                 collection_name_context: str,
                 task_step_store: BaseTaskStepStore,
                 task_step_to_question_chain: Chain,
                 task_step_question_to_graphql_chain: Chain,
                 client: WeaviateClient,
                 cross_encoder: CrossEncoder):
        self.base_path = base_path
        self.start_task_context = start_task_context
        self.collection_name = collection_name
        self.collection_name_context = collection_name_context
        self.task_step_store = task_step_store
        self.task_step_to_question_chain = task_step_to_question_chain
        self.task_step_question_to_graphql_chain = task_step_question_to_graphql_chain
        self.client = client
        self.cross_encoder = cross_encoder

    @classmethod
    def from_task_step_to_question_chain(
            cls,
            base_path: str,
            start_task_context: str,
            llm_runable: Runnable[LanguageModelInput, BaseMessage],
            task_step_store: BaseTaskStepStore,
            client: WeaviateClient,
            cross_encoder_path: str
    ) -> TaskStepToQuestionChain:
        """
        
        1、对当前主任务名称创建hash值，作为collection_name
        2、对当前任务步骤名称创建hash值，作为collection_name_context
        """
        prompt_template1 = PromptTemplate(input_variables=["start_task_context", 
                                                           "aemo_representation_context",
                                                           "task_step_name",
                                                           "task_step_description",
                                                           "task_step_level"],
                                          template=os.environ.get(
                                "CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE", CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE
                            ))

        task_step_to_question_chain = (prompt_template1 | llm_runable | StrOutputParser())
 
        def wrapper_output(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_context": _dict["task_step_question_context"],
            }

        task_step_to_question_chain = ({
                                        "task_step_question_context": task_step_to_question_chain,
                                    }
                                    | RunnableLambda(wrapper_output))

        prompt_template2 = PromptTemplate(input_variables=[
                                            "collection_name_context", 
                                            "task_step_question"
                                        ],
                                        template=os.environ.get("TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE", TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE)
                                        )

        task_step_question_to_graphql_chain = (prompt_template2 | llm_runable | StrOutputParser())

        def wrapper_output2(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_graphql_context": _dict["task_step_question_graphql_context"],
            }

        task_step_question_to_graphql_chain = ({
                                        "task_step_question_graphql_context": task_step_question_to_graphql_chain,
                                    }
                                    | RunnableLambda(wrapper_output2))

        cross_encoder = CrossEncoder(
            cross_encoder_path,
            automodel_args={"torch_dtype": "auto"},
            trust_remote_code=True,
        )
        
           
        collection_id = get_query_hash(start_task_context)
        # TODO 之后单独设计召回模块，由模块生成collection_name, collection_name_context
        collection_name, collection_name_context = init_context_collections(client, collection_id)
        return cls(
            base_path=base_path,
            start_task_context=start_task_context,
            collection_name=collection_name,
            collection_name_context=collection_name_context,
            task_step_store=task_step_store,
            task_step_to_question_chain=task_step_to_question_chain,
            task_step_question_to_graphql_chain=task_step_question_to_graphql_chain,
            client=client,
            cross_encoder=cross_encoder
        )

    def invoke_task_step_to_question(self, task_step_id: str) -> None:
        """
        对开始任务进行抽取，得到任务步骤
        :return:
        """ 

        task_step_node = self.task_step_store.get_task_step(task_step_id)
        result = self.task_step_to_question_chain.invoke(task_step_node.__dict__)
        task_step_node.task_step_question = result["task_step_question_context"]
        self.task_step_store.add_task_step([task_step_node])
        
        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)
        self.task_step_store.persist(persist_path=task_step_store_path) 


    def _retrieve_task_step_question_context(self, task_step_id: str, collection_name_context: str, task_step_node: TaskStepNode) -> tuple[List[str], str, str]:
        """基于概念路径检索方法"""

        result = self.task_step_question_to_graphql_chain.invoke({
            "collection_name_context": collection_name_context,
            "task_step_question": task_step_node.task_step_question
        })
        # 解析出graphql查询语句
        action_match = PATTERN.search(result["task_step_question_graphql_context"])
        if action_match is not None:
            # 去掉前后空格
            temp_graphql_context = action_match.group(1).strip()
            # 去掉 ```graphql , 后```
            task_step_question_graphql_context = temp_graphql_context.replace("```graphql", "").replace("```", "")
        else:
            raise ValueError("No graphql query found in the response")

        response = self.client.graphql_raw_query(task_step_question_graphql_context)
        data_list = response.get[collection_name_context]
        data_list_copy = copy.deepcopy(data_list)
                
        # 找出根节点
        root_nodes = find_root_nodes(data_list_copy)
        logger.info(f"Root nodes:{json.dumps(root_nodes)}")

        # 找出出度大于3的节点
        high_outdegree_concepts = find_high_outdegree_concepts(data_list_copy)
        logger.info(f"Concepts with more than 3 outgoing edges:{high_outdegree_concepts}")


        # 找出入度大于3的节点
        high_indegree_concepts = find_high_indegree_nodes(data_list_copy)
        logger.info(f"Concepts with more than 3 going edges:{high_indegree_concepts}")


        create_interactive_graph(
            data_list_copy,
            root_nodes_with_concept=root_nodes,
            high_outdegree_concepts=high_outdegree_concepts,
            high_indegree_concepts=high_indegree_concepts,
            filename=f"{self.base_path}/storage/{task_step_id}_semantic_path_interactive.html"
        )

        G = create_G(data_list_copy,  root_nodes_with_concept=root_nodes,
            high_outdegree_concepts=high_outdegree_concepts,
            high_indegree_concepts=high_indegree_concepts)

        result_paths = find_simple_path(G,  root_nodes_with_concept=root_nodes,
            high_outdegree_concepts=high_outdegree_concepts)

        logger.info(f"Root nodes:{json.dumps(root_nodes)}")
        
        logger.info(f"Concepts with more than 3 outgoing edges:{high_outdegree_concepts}")
        
        logger.info(f"Concepts with more than 3 going edges:{high_indegree_concepts}")

        # result_paths 中即为符合条件的路径列表
        refIds = []
        if len(result_paths) == 0:
            raise ValueError("No paths found")
        # 输出路径以及路径中节点的refId信息
        for p in result_paths:
            logger.info(f"Path: " + " -> ".join(p))
            # 输出路径中每个节点的refId字段信息
            for node in p:
                logger.info(f"Node: {node}, refIds: {G.nodes[node].get('refId')}")
                refIds.extend(G.nodes[node].get('refId'))

        return refIds

    def invoke_task_step_question_context(self, task_step_id: str) -> None:
        """
        对任务步骤进行抽取，得到任务步骤的上下文
        3、增加项目
        """

        task_step_node = self.task_step_store.get_task_step(task_step_id)
        if task_step_node.task_step_question_context is not None and len(task_step_node.task_step_question_context) > 0:
            return

        top_k =100  # 可选参数，默认查询返回前 100个结果
        properties_list = exe_query(task_step_node.task_step_question,top_k)
        # 插入数据到数据库 
        insert_into_database(self.client, self.collection_name, 'ref_id', properties_list)

        context_properties_list = [{
            "refId": item.get('ref_id'),
            "paperId": item.get('paper_id'),
            "chunkId": item.get('chunk_id'),
            "chunkText": item.get('chunk_text')
        } for item in properties_list]

        insert_into_database(self.client, self.collection_name_context, 'refId', context_properties_list)
        retrieve_ids = []
        try:
            retrieve_ids = self._retrieve_task_step_question_context(task_step_id, self.collection_name_context, task_step_node)
        except Exception as e:
            logger.error(f"Error retrieving task step question context: {e}")
            
        if len(retrieve_ids) == 0:
            
            collection = self.client.collections.get(self.collection_name) 

            vector_names = ["chunk_text"]
            response = collection.query.hybrid(
                query=task_step_node.task_step_question,  
                target_vector=vector_names,  # Specify the target vector for named vector collections
                limit=1,
                alpha=0.1,
                query_properties=["chunk_text"],  
                return_metadata=MetadataQuery(score=True, explain_score=True, distance=True),
            )
        else:
            # 对refIds进行编码
            collection = self.client.collections.get(self.collection_name)
            response = collection.query.fetch_objects(
                filters=Filter.by_property("ref_id").contains_any(retrieve_ids), 
                return_metadata=MetadataQuery(score=True, explain_score=True, distance=True),
            )

        chunk_texts = []
        ref_ids = []
        for o in response.objects:  
            chunk_texts.append(o.properties['chunk_text'])
            ref_ids.append(o.properties['ref_id'])
        
        rankings = self.cross_encoder.rank(task_step_node.task_step_question, chunk_texts, return_documents=True, convert_to_tensor=True)
                    
        task_step_question_context = []
        # 召回问题前3条，存入task_step_question_context
        for i, ranking in enumerate(rankings[:3]):
            logger.info(f"ref_ids: {ref_ids[i]}, Score: {ranking['score']:.4f}, Text: {ranking['text']}")
            task_step_question_context.append(TaskStepContext(ref_id=ref_ids[i], score=ranking['score'], text=ranking['text']))

        # rankingscorpus_id 的索引与ref_ids的索引一致 
        task_step_node.task_step_question_context = task_step_question_context
        self.task_step_store.add_task_step([task_step_node])

        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)
        self.task_step_store.persist(persist_path=task_step_store_path) 


    def export_csv_file_path(self, task_step_id: str) -> str:
        """
        3、对召回内容与问题 导出csv文件
        """ 
        task_step_node = self.task_step_store.get_task_step(task_step_id)
            
        table_data = []  
        row = [
            task_step_id,
            task_step_node.task_step_name,
            task_step_node.task_step_level
        ]
        
        table_data.append(row) 
        row2 = [
            task_step_id,
            task_step_node.task_step_question,
            task_step_node.task_step_level
        ]
        table_data.append(row2)
        for context in task_step_node.task_step_question_context:
            row3 = [
                task_step_id,
                f"ref_ids: {context['ref_id']}, Score: {context['score']:.4f}, Text: {context['text']}",
                task_step_node.task_step_level
            ]
            table_data.append(row3)

        table = pd.DataFrame(table_data, columns=["角色", "内容", "分镜"])

        table.to_csv(f"{self.base_path}/storage/{task_step_id}.csv", index=False)

        return f"{self.base_path}/storage/{task_step_id}.csv"


import os
import time

from llama_index.llms.openai import OpenAI
from llama_index.embeddings.openai import (
    OpenAIEmbedding,
)  # pants: no-infer-dep

from llama_index.llms.openai import OpenAI
from llama_index.finetuning.callbacks import OpenAIFineTuningHandler
from llama_index.core.callbacks import CallbackManager
from llama_index.core.agent import ReActAgent

from llama_index.core import Settings

from dreamsboard.agent.react_agent_build_query_engine_tools import build_query_docs, build_query_docs_index_store, \
    build_query_engine_tools
from dreamsboard.agent.react_tools_utils import save_questions, load_questions
from llama_index.llms.openai.utils import GPT4_MODELS


if __name__ == '__main__':

    llm = OpenAI(model="glm-4", temperature=0.99, api_key="a9733a59370d34ef51d261bd461251a5.YD9PU7sAqUwgD8Sk",
                 api_base="https://open.bigmodel.cn/api/paas/v4/")
    embeddings = OpenAIEmbedding(api_key="EMPTY",  model="text-embedding-3-large", api_base="http://127.0.0.1:9997/v1")
    query_docs = build_query_docs()
    index_store = build_query_docs_index_store(query_docs, embeddings)
    query_engine_tools = build_query_engine_tools(llm, index_store)
    # 获取当前文件路径
    root_dir = os.path.dirname(os.path.abspath(__file__))
    train_questions = load_questions(os.path.join(root_dir, f"glm_train_questions_1152q.txt"))
    eval_questions = load_questions(os.path.join(root_dir, f"glm_eval_questions_288q.txt"))
    print(len(eval_questions))

    finetuning_handler = OpenAIFineTuningHandler()
    callback_manager = CallbackManager([finetuning_handler])
    # limit the context window artifically to test refine process
    Settings.context_window = 4096
    agent_llm = OpenAI(model="glm-4", temperature=0.01, api_key="a9733a59370d34ef51d261bd461251a5.YD9PU7sAqUwgD8Sk",
                       api_base="https://open.bigmodel.cn/api/paas/v4/")
    gpt4_agent = ReActAgent.from_tools(
        query_engine_tools,
        llm=agent_llm,
        callback_manager=callback_manager,
        verbose=True,
    )
    # 增加一个时间戳
    # train_questions = [f"{q} {time.time()}" for q in train_questions]
    # 每10次迭代保存一次模型
    for idx, question in enumerate(train_questions):
        try:
            print(f"[{idx}] Question: {question}")
            response = gpt4_agent.query(question)
            print(f"[{idx}] Agent Response: {str(response)}")
            if idx % 10 == 0:
                # save events
                print(f"[{idx}] Saving finetuning events...")
                finetuning_handler.save_finetuning_events(os.path.join(root_dir, f"glm_q_gemini_a_finetuning_events_{idx}q_{time.time()}.jsonl"))
        except Exception as e:
            print(f"[{idx}] Error: {e}")
    # save events
    finetuning_handler.save_finetuning_events(os.path.join(root_dir, f"glm_q_gemini_a_finetuning_events_{len(train_questions)}q_{time.time()}.jsonl"))



from dreamsboard.engine.generate.code_executor import CodeExecutor
from dreamsboard.engine.generate.code_generate import CodeGenerator


# 创建一个责任链节点
class CodeGeneratorHandler:
    def __init__(self, generator: CodeGenerator, next_handler=None):
        self.generator = generator
        self.next_handler = next_handler

    def generate(self, render_data: dict):
        code = self.generator.generate(render_data)
        if self.next_handler:
            code += self.next_handler.generate(render_data)
        return code


# Create a code generator builder
class CodeGeneratorChain:
    def __init__(self):
        self.chain_head: CodeGeneratorHandler = None
        self.chain_tail: CodeGeneratorHandler = None  # Track the tail of the chain

    def add_generator(self, generator: CodeGenerator):
        handler = CodeGeneratorHandler(generator)
        if self.chain_head is None:
            self.chain_head = handler
            self.chain_tail = handler
        else:
            self.chain_tail.next_handler = handler  # Update the tail
            self.chain_tail = handler  # Update the new tail

    def remove_last_generator(self):
        if self.chain_head is None:
            return

        if self.chain_head == self.chain_tail:
            # Only one handler in the chain
            self.chain_head = None
            self.chain_tail = None
        else:
            current_handler = self.chain_head
            while current_handler.next_handler != self.chain_tail:
                current_handler = current_handler.next_handler
            current_handler.next_handler = None
            self.chain_tail = current_handler

    def generate(self, render_data: dict):
        if self.chain_head is None:
            raise RuntimeError("chain_head is None.")

        return self.chain_head.generate(render_data)



from __future__ import annotations
from typing import List, Optional
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from dreamsboard.document_loaders.ner_loader import NerLoader
from dreamsboard.document_loaders.protocol.ner_protocol import DreamsStepInfo, Personality
from dreamsboard.engine.engine_builder import CodeGeneratorBuilder
from dreamsboard.engine.generate.code_generate import (
    BaseProgramGenerator,
    QueryProgramGenerator,
    AIProgramGenerator,
    EngineProgramGenerator,
)
from dreamsboard.document_loaders import StructuredStoryboardCSVBuilder, KorLoader
from langchain.chains import LLMChain
import logging

from dreamsboard.engine.loading import load_store_from_storage
from dreamsboard.engine.storage.storage_context import StorageContext

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class StructuredDreamsStoryboard:
    """
   
    构建会话场景执行器 StructuredDreamsStoryboard
        对剧本和分析结果进行结构化，将开放问题与性格分析结果进行结合。生成情景扮演会话器
    此过程如下
        对开放问题结果进行抽取，得到问题内容
        对性格分析结果进行抽取，得到性格分析结果
        增加系统提示词
        根据剧本与任务性格基础情景扮演代码，根据每步的抽取析得到的问题，生成问题的答案
        在上下文中增加，关于人物职业的引导话术
        导出情景扮演会话器
    """

    def __init__(self,
                 builder: StructuredStoryboardCSVBuilder,
                 dreams_guidance_context: str,
                 dreams_personality_context: str,
                 kor_dreams_guidance_chain: LLMChain,
                 kor_dreams_personality_chain: LLMChain,
                 ner_dreams_personality_chain: LLMChain,
                 user_id: str = None,
                 llm_runable: Runnable[LanguageModelInput, BaseMessage]|None = None,
                 ):
        """

        :param builder: 剧本
        :param dreams_guidance_context: 开放问题
        :param dreams_personality_context: 性格分析结果
        """
        self.builder = builder
        self.dreams_guidance_context = dreams_guidance_context
        self.dreams_personality_context = dreams_personality_context
        self.kor_dreams_guidance_chain = kor_dreams_guidance_chain
        self.kor_dreams_personality_chain = kor_dreams_personality_chain
        self.ner_dreams_personality_chain = ner_dreams_personality_chain
        self.user_id = user_id
        self.llm_runable = llm_runable
        
    @classmethod
    def form_builder(cls,
                     llm_runable: Runnable[LanguageModelInput, BaseMessage],
                     builder: StructuredStoryboardCSVBuilder,
                     dreams_guidance_context: str,
                     dreams_personality_context: str,
                     guidance_llm: Runnable[LanguageModelInput, BaseMessage] = None,
                     personality_llm: Runnable[LanguageModelInput, BaseMessage] = None,
                     user_id: str = None, ) -> StructuredDreamsStoryboard:
        kor_dreams_guidance_chain = KorLoader.form_kor_dreams_guidance_builder(
            llm_runable=llm_runable if guidance_llm is None else guidance_llm)
        kor_dreams_personality_chain = KorLoader.form_kor_dreams_personality_builder(
            llm_runable=llm_runable if personality_llm is None else personality_llm
        )
        ner_dreams_personality_chain = NerLoader.form_ner_dreams_personality_builder(
            llm_runable=llm_runable if personality_llm is None else personality_llm
        )

        return cls(builder=builder,
                   dreams_guidance_context=dreams_guidance_context,
                   dreams_personality_context=dreams_personality_context,
                   kor_dreams_guidance_chain=kor_dreams_guidance_chain,
                   kor_dreams_personality_chain=kor_dreams_personality_chain,
                   ner_dreams_personality_chain=ner_dreams_personality_chain,
                   user_id=user_id,
                   llm_runable=llm_runable)

    def kor_dreams_guidance_context(self) -> List[DreamsStepInfo]:
        """
        对开放问题结果进行抽取，得到问题内容
        :return:
        """
        response = self.kor_dreams_guidance_chain.run(self.dreams_guidance_context)
        dreams_step_list = []
        if response.get('data') is not None and response.get('data').get('script') is not None:
            step_list = response.get('data').get('script')
            for step in step_list:
                dreams_step = DreamsStepInfo(step_advice=step.get('step_advice'),
                                             step_description=step.get('step_description'))
                dreams_step_list.append(dreams_step)

        return dreams_step_list

    def kor_dreams_personality_context(self) -> str:
        """
        对性格分析结果进行抽取，得到性格分析结果
        :return:
        """
        response = self.kor_dreams_personality_chain.run(self.dreams_personality_context)

        personality = ""
        if response.get('data') is not None and response.get('data').get('personality_script') is not None:
            personality_list = response.get('data').get('personality_script')
            # [{'personality': '具有情感表达和期待、注重个体快感、善于运用语义信息、对社会行为产生兴趣'}]},
            # 拼接personality 成一个字符串
            for item in personality_list:
                personality += item.get('personality') + "、"

        return personality

    def ner_dreams_personality_context(self) -> str:
        """
        对性格分析结果进行抽取，得到性格分析结果,ner版本
        :return:
        """
        response: Personality = self.ner_dreams_personality_chain.invoke({"input": self.dreams_personality_context})

        return response.personality

    def builder_base_cosplay_code(self) -> List[str]:
        """
        根据剧本与任务性格基础情景扮演代码，
        :return:
        """

        self.builder.load()
        # 组织内容和角色
        messages = []
        for storyboard in self.builder.data:
            role = storyboard.story_board_role.name
            text = storyboard.story_board_text.name
            if len(role) > 0 and len(text) > 0:
                messages.append(f"{role}:「{''.join(text)}」\n")
            else:
                messages.append(f"「{''.join(text)}」\n")

        return messages

    def loader_cosplay_builder(self, 
                               dreams_cosplay_role: str = "心理咨询工作者", 
                               dreams_cosplay_step: int = 1, 
                               storage_context: Optional[StorageContext] = None,
                               ) -> CodeGeneratorBuilder:
        """
        dreams_cosplay_role对话角色
        dreams_cosplay_step对话次数
        """
        code_gen_builder = CodeGeneratorBuilder.from_template(nodes=[], storage_context=storage_context)

        # 创建一个字典，用于按照story_board组织内容和角色
        export_role = self.builder.export_role()
        if self.user_id is None:
            # 获取export_role set的第一个值
            cosplay_role = export_role.pop()
        else:
            cosplay_role = self.user_id

        guidance_questions = self.kor_dreams_guidance_context()
        try:
            personality = self.ner_dreams_personality_context()
        except Exception as e:
            logger.error(f'ner_dreams_personality_context error:{e}')
            personality = self.kor_dreams_personality_context()

        base_cosplay_message = self.builder_base_cosplay_code()
        _base_render_data = {
            'cosplay_role': cosplay_role,
            'personality': personality,
            'messages': base_cosplay_message
        }
        code_gen_builder.add_generator(BaseProgramGenerator.from_config(cfg={
            "code_file": "base_template.py-tpl",
            "render_data": _base_render_data,
        }))

        for index, guidance_question in enumerate(guidance_questions):
            if dreams_cosplay_step >= index:
                break
            
            logger.info(f'{guidance_question.step_description}:{guidance_question.step_advice}')
            _dreams_render_data = {
                'dreams_cosplay_role': dreams_cosplay_role,
                'dreams_message': guidance_question.step_advice,
            }
            code_gen_builder.add_generator(QueryProgramGenerator.from_config(cfg={
                "query_code_file": "dreams_query_template.py-tpl",
                "render_data": _dreams_render_data,
            })) 
            executor = code_gen_builder.build_executor(
                llm_runable=self.llm_runable,
                messages=[]
            )
            executor.execute()
            _ai_message = executor.chat_run()

            logger.info(f'{guidance_question.step_description}:{_ai_message}')

            _ai_render_data = {
                'ai_message_content': _ai_message.content
            }
            code_gen_builder.add_generator(AIProgramGenerator.from_config(cfg={
                "ai_code_file": "ai_template.py-tpl",
                "render_data": _ai_render_data,
            }))

        return code_gen_builder



from weaviate.classes.query import Filter
from weaviate.client import WeaviateClient
from typing import List

import requests

import hashlib
import os
import json
import tempfile 
import logging 

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)
CACHE_DIR = os.path.join(tempfile.gettempdir(), "query_cache")  # 自定义缓存目录
# 定义接口 URL
SEARCH_PAPERS_URL = "http://180.184.65.98:38880/atomgit/search_papers"
QUERY_BY_PAPER_ID_URL = "http://180.184.65.98:38880/atomgit/query_by_paper_id"


def prepare_properties(paper_details: List[dict]):
    """
    将 paper_details 数据转化为向量数据库插入所需的 properties 格式
    """
    properties_list = []
    
    for paper in paper_details:
        properties = {
            "ref_id": paper.get("paper_id"),  # 可能是另一个 ID，具体根据实际情况
            "paper_id": paper.get("paper_id"),
            "paper_title": paper.get("paper_title"),
            "chunk_id": paper.get("chunk_id"),
            "chunk_text": paper.get("chunk_text"),
            "original_filename": paper.get("original_filename", "")  # 默认空字符串，如果没有提供
        }
        properties_list.append(properties)
    
    return properties_list

def insert_into_database(client: WeaviateClient, collection_name:str, union_id:str, properties_list: List[dict]):
    """
    插入数据到向量数据库,检查唯一
    """
    
    collection = client.collections.get(collection_name)
    union_ids = [item.get(union_id) for item in properties_list]
    response = collection.query.fetch_objects(
        filters=Filter.by_property(union_id).contains_any(union_ids),
    )
        
    exist_ids = [o.properties[union_id] for o in response.objects]
   
    for properties in properties_list:
        if properties[union_id] not in exist_ids:
            collection.data.insert(properties=properties)
    

def search_papers(query, top_k=5):
    """
    根据查询文本 search_papers 接口进行模糊查询，返回论文 ID 列表
    """
    params = {'query': query, 'top_k': top_k}
    response = requests.get(SEARCH_PAPERS_URL, params=params)
    
    if response.status_code == 200:
        # 返回的 JSON 数据，假设是一个包含论文信息的数组
        return response.json()  # 返回论文的列表
    else:
        logger.info(f"Error: {response.status_code}")
        return None

def query_by_paper_id(paper_id, top_k=5):
    """
    根据 paper_id 调用 query_by_paper_id 接口，获取该论文的详细信息
    """
    params = {'paper_id': paper_id, 'top_k': top_k}
    response = requests.get(QUERY_BY_PAPER_ID_URL, params=params)
    
    if response.status_code == 200:
        # 返回的 JSON 数据，假设是论文的详细信息
        return response.json()  # 返回论文的详细信息
    else:
        logger.info(f"Error: {response.status_code}")
        return None
 


def ensure_cache_dir():
    """确保缓存目录存在"""
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)

def get_query_hash(query):
    """根据 query 生成唯一的哈希值"""
    return hashlib.md5(query.encode('utf-8')).hexdigest()

def get_cache_file_path(query):
    """获取 query 对应的缓存文件路径"""
    query_hash = get_query_hash(query)
    return os.path.join(CACHE_DIR, f"{query_hash}.json")

def check_cache(query):
    """检查是否存在 query 的缓存文件"""
    file_path = get_cache_file_path(query)
    if os.path.exists(file_path):
        logger.info(f"Cache hit file_path: {file_path} for query: {query}")
        with open(file_path, 'r', encoding='utf-8') as cache_file:
            return json.load(cache_file)  # 返回缓存内容
    return None

 
def save_to_cache(query, data):
    """将数据保存到 query 的缓存文件中"""
    ensure_cache_dir()
    file_path = get_cache_file_path(query)
    with open(file_path, 'w', encoding='utf-8') as cache_file:
        json.dump(data, cache_file, ensure_ascii=False, indent=4)
        logger.info(f"Data cached for query: {query} at {file_path}")



def exe_query(query, top_k):
    cached_data = check_cache(query)
    if cached_data:
        logger.info(f"Using cached data for query: {query}")
        return cached_data  # 如果有缓存则直接返回
    # 第一步：模糊查询论文
    logger.info(f"Searching papers for query: {query}")
    papers = search_papers(query, top_k)
    
    if papers:
        logger.info(f"Found {len(papers)} papers.")
        
        # 获取所有的 paper_id，并去重
        paper_ids = set()  # 使用 set 去重
        for paper in papers:
            paper_id = paper.get("entity", {}).get("paper_id")
            if paper_id:
                paper_ids.add(paper_id)  # 添加到 set 中，自动去重
        
        # 输出去重后的 paper_id 数量
        logger.info(f"Found {len(paper_ids)} unique paper IDs.")
    
         # 查询每个 unique paper_id 的详细信息
        all_paper_details = [] 
        for paper_id in paper_ids:
            logger.info(f"Fetching details for paper ID: {paper_id}")
            paper_details = query_by_paper_id(paper_id, top_k)
            if paper_details:
                all_paper_details.extend(paper_details)  # 假设返回的是一个列表
            else:
                logger.info(f"Failed to fetch details for paper ID: {paper_id}")
        
        # 将获取的 paper_details 转换为向量数据库可插入的格式
        properties_list = prepare_properties(all_paper_details)
        # 保存 properties_list 到临时文件
        save_to_cache(query, properties_list)
        return properties_list
    else:
        logger.info("No papers found.")
     


# Import necessary libraries
import matplotlib.pyplot as plt
import networkx as nx 
from matplotlib import rcParams
import os
import copy

from pyvis.network import Network
from IPython.core.display import display, HTML
import networkx as nx
import json
import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)
# 设置 HTTP 和 HTTPS 代理
# os.environ["HTTP_PROXY"] = "http://127.0.0.1:7890"
# os.environ["HTTPS_PROXY"] = "http://127.0.0.1:7890"


# 设置中文字体为 SimHei（黑体）
rcParams['font.sans-serif'] = ['DejaVu Serif']  # Example
 
# rcParams['axes.unicode_minus'] = False  # 解决负号显示问题

# Function to create a graph from semantic path data
def create_caption_graph(data):
    G = nx.DiGraph()
    for i, node in enumerate(data):
        concept = node["concept"]
        G.add_node(concept, distanceToQuery=node["distanceToQuery"], distanceToResult=node["distanceToResult"])
        if i < len(data) - 1:
            next_concept = data[i + 1]["concept"]
            G.add_edge(concept, next_concept, weight=node["distanceToNext"])
    return G

# 图形绘制函数，带上下翻译对比
def plot_graph_with_caption(G, title, caption, pos=None, translate=True):
    translated_title, translated_caption = title, caption
    
    if translate:
        # 自动翻译标题和说明
        translated_title = title
        # translated_caption = GoogleTranslator(source="en", target="zh-CN").translate(caption)
        translated_caption = caption

    plt.figure(figsize=(10, 10))
    pos = pos or nx.spring_layout(G, seed=42)  # 节点布局
    nx.draw(
        G, pos, with_labels=True, node_color="lightblue", edge_color="gray",
        node_size=2000, font_size=10
    )
    labels = nx.get_edge_attributes(G, 'weight')
    nx.draw_networkx_edge_labels(G, pos, edge_labels={k: f"{v:.3f}" for k, v in labels.items()})

    # 显示翻译和原文对比
    plt.title(f"{translated_title}\n(Original: {title})", fontsize=14, pad=20)
    plt.figtext(
        0.5, 0.01,
        f"Translated Caption: {translated_caption}\n(Original Caption: {caption})",
        wrap=True, horizontalalignment='center', fontsize=12
    )
    plt.show()

     
  

# Updated caption graph function
def create_and_plot_all(data_list):
    for i, context in enumerate(data_list):
        semantic_path = context["_additional"]["semanticPath"]["path"]
        caption = context["chunkText"]
        graph = create_caption_graph(semantic_path)
        title = f"Semantic Path Visualization {i+1}"
        plot_graph_with_caption(graph, title, caption)



 
# Function to find root nodes and their properties
def find_root_nodes(data):
    # Create a directed graph using NetworkX
    G = nx.DiGraph()
    node_properties = []

    for context in data: 
        path = context["_additional"]["semanticPath"]["path"]
        for i, node in enumerate(path):
            concept = node["concept"]
             
            # Add node with its properties
            if concept not in G.nodes:
                G.add_node(concept)
            if i < len(path) - 1:
                next_concept = path[i + 1]["concept"]
                if next_concept not in G.nodes:
                    G.add_node(next_concept) 
                G.add_edge(concept, next_concept)

    # Find nodes with in-degree 0 (root nodes)
    root_nodes = [node for node, in_degree in G.in_degree() if in_degree == 0]
      
    return root_nodes

# Function to find concepts with more than 3 outgoing edges
def find_high_outdegree_concepts(data, edge_threshold=3):
    # Create a directed graph using NetworkX
    G = nx.DiGraph()

    for context in data:
        path = context["_additional"]["semanticPath"]["path"]
        for i, node in enumerate(path):
            concept = node["concept"]
            if concept not in G.nodes:
                G.add_node(concept)
            if i < len(path) - 1:
                next_concept = path[i + 1]["concept"]
                if next_concept not in G.nodes:
                    G.add_node(next_concept)
                G.add_edge(concept, next_concept)

    # Find nodes with outgoing edges greater than the threshold
    high_outdegree_concepts = [node for node, out_degree in G.out_degree() if out_degree > edge_threshold]
    return high_outdegree_concepts


# Function to find nodes with in-degree greater than a threshold
def find_high_indegree_nodes(data, threshold=3):
    # Create a directed graph using NetworkX
    G = nx.DiGraph() 
    for context in data: 
        path = context["_additional"]["semanticPath"]["path"]
        for i, node in enumerate(path):
            concept = node["concept"]
 
            # Add node with its properties
            if concept not in G.nodes:
                G.add_node(concept)
            if i < len(path) - 1:
                next_concept = path[i + 1]["concept"]
                if next_concept not in G.nodes:
                    G.add_node(next_concept)
                G.add_edge(concept, next_concept)
 
    # Find nodes with in-degree greater than the threshold
    high_indegree_nodes = [
        node for node, in_degree in G.in_degree() if in_degree > threshold
    ]
    return high_indegree_nodes

def get_color(concept, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts):
    # 高亮根节点
    if concept in root_nodes_with_concept:
        color = "green"  # 绿色表示根节点
    # 高亮出度较大的节点
    elif concept in high_outdegree_concepts:
        color = "red"  # 红色表示出度较大的节点
    elif concept in high_indegree_concepts:
        color = "blue"  # 蓝色表示出度较大的节点
    else:
        color = "lightblue"  # 默认节点颜色 

    return color
        
# 修改 create_interactive_graph 函数以标注特定节点
def create_interactive_graph(data, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts, filename="semantic_path_interactive.html"):
    net = Network(height="750px", width="100%", directed=True)
 
    for context in data:
        # 直接使用原始路径数据
        path = context["_additional"]["semanticPath"]["path"]
        nodes_added = set()  # Track added nodes
        for i, node in enumerate(path):
            concept = node["concept"]

           
            if concept not in nodes_added:
                color = get_color(concept, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts) 
                net.add_node(concept, label=concept, color=color)
                nodes_added.add(concept)

            if i < len(path) - 1:
                next_concept = path[i + 1]["concept"]
                # Ensure the next concept exists before adding the edge
                if next_concept not in nodes_added:
                    
                    color = get_color(next_concept, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts) 
                    net.add_node(next_concept, label=next_concept, color=color)
                    nodes_added.add(next_concept)
                net.add_edge(concept, next_concept, title=f"Distance: {node['distanceToNext']:.3f}")

    # 保存 HTML 文件
    net.write_html(filename)

    # 在 Jupyter 中显示 HTML
    # iframe = f'<iframe src="{filename}" width="100%" height="750px" frameborder="0"></iframe>'
    # display(HTML(iframe))
     


# Function to find concepts with more than 3 outgoing edges
def create_G(data, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts):
    # 创建有向图
    G = nx.DiGraph()
 
    for context in data:
        path = context["_additional"]["semanticPath"]["path"]
        refId = context.get("refId", None)  # 从context中获取refId字段
        nodes_added = set()  # 用于在同一个context中避免重复添加同一节点属性
        
        for i, node in enumerate(path):
            concept = node["concept"]
            
            # 如果节点在全局图中不存在，则新建节点并添加属性
            if concept not in G:
                color = get_color(concept, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts)
                # 将refId信息存入节点属性中。如果同一节点在多个context中出现，可以存成列表/集合
                G.add_node(concept, label=concept, color=color, refId=[refId] if refId else [])
            else:
                # 如果节点已存在，且有新的refId信息，合并进去（去重）
                if refId and refId not in G.nodes[concept]['refId']:
                    G.nodes[concept]['refId'].append(refId)
            
            nodes_added.add(concept)

            if i < len(path) - 1:
                next_concept = path[i + 1]["concept"]
                if next_concept not in G:
                    color = get_color(next_concept, root_nodes_with_concept, high_outdegree_concepts, high_indegree_concepts)
                    G.add_node(next_concept, label=next_concept, color=color, refId=[refId] if refId else [])
                else:
                    # 如果下一个概念节点已存在，更新refId信息
                    if refId and refId not in G.nodes[next_concept]['refId']:
                        G.nodes[next_concept]['refId'].append(refId)

                # 增加边的属性（这里已有Distance信息）
                G.add_edge(concept, next_concept, title=f"Distance: {node['distanceToNext']:.3f}")

    return G



def find_simple_path(G, root_nodes_with_concept, high_outdegree_concepts):
        
    # 筛选出那些包含红色节点的路径
    result_paths = []
    for root in root_nodes_with_concept:
        for out in high_outdegree_concepts:
            if nx.has_path(G, root, out):
                # 寻找从 "root" 到 "out" 的所有简单路径
                paths = nx.all_shortest_paths(G, source=root, target=out)
                
                for path in paths:
                    # 判断路径中是否存在 color='red' 的节点
                    if any(G.nodes[node].get("color") == "red" for node in path):
                        if len(path) == 3:
                            result_paths.append(path)
            else:
                logger.info(f"No path between {root} and {out}")
                continue
                
    return result_paths




"""
使用设计模式，构建一个代码生成器，功能如下，
程序动态的生成python代码，python代码有三种形式，1、基础程序。2、逻辑控制程序。3、逻辑加载程序。4、逻辑运行程序

程序的执行流程如下：
1、
构建出的代码通过exec函数运行

2、
构建出的代码最终通过建造者生成一个执行器

3、
建造者在执行CodeGenerator的时候，需要使用责任链实现


"""

from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Any, Optional, Dict

from jinja2 import Template
import logging

import textwrap
import hashlib
from pydantic import Field

from dreamsboard.engine.schema import BaseNode, ObjectTemplateType, TRUNCATE_LENGTH, WRAP_WIDTH
from dreamsboard.templates import get_template_path

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


def truncate_text(text: str, max_length: int) -> str:
    """Truncate text to a maximum length."""
    if len(text) <= max_length:
        return text
    return text[: max_length - 3] + "..."


# 创建一个代码生成器抽象类
class CodeGenerator(BaseNode, ABC):

    @classmethod
    def from_config(cls, cfg=None):
        return cls()

    @property
    @abstractmethod
    def render_data(self) -> dict:
        """Get render_data."""

    @render_data.setter
    @abstractmethod
    def render_data(self, _render_data: dict) -> None:
        """Get render_data."""

    @property
    @abstractmethod
    def render_code(self) -> str:
        """Get render_code."""

    @render_code.setter
    @abstractmethod
    def render_code(self, _exec_code: str) -> None:
        """Get render_code."""

    def generate(self, render_data: dict = {}) -> str:
        base_template = Template(self.template_content)
        if render_data is not None and self.render_data is not None:
            self.render_data = {**render_data, **self.render_data}
        else:
            # 处理其中一个或两者都为 None 的情况
            self.render_data = render_data or self.render_data or {}

        self.render_code = base_template.render(self.render_data)

        return self.render_code

    def calculate_md5(self):
        md5_hash = hashlib.md5()
        md5_hash.update(self.render_code.encode('utf-8'))
        return md5_hash.hexdigest()

    def __str__(self) -> str:
        source_text_truncated = truncate_text(
            self.render_code.strip(), TRUNCATE_LENGTH
        )
        source_text_wrapped = textwrap.fill(
            f"Text: {source_text_truncated}\n", width=WRAP_WIDTH
        )
        return f"Node ID: {self.node_id}\n{source_text_wrapped}"


# 创建不同类型的代码生成器
class BaseProgramGenerator(CodeGenerator):
    exec_code: Optional[str] = Field(
        default="", description="执行代码"
    )
    base_template_content: Optional[str] = Field(
        default="", description="模板内容",
    )
    exec_data: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="执行环境数据",
        alias="exec_data",
    )

    def __init__(self, code_file: str = None, render_data=None, **kwargs):
        # 检查kwargs中是否包含这些属性，如果包含就不执行文件读取
        if "exec_data" not in kwargs and "base_template_content" not in kwargs and "exec_code" not in kwargs:
            if render_data is None:
                render_data = {}
            # 读取模板文件
            with open(code_file, 'r', encoding='utf-8') as template_file:
                base_template_content = template_file.read()

            super().__init__(exec_data=render_data, base_template_content=base_template_content)
        else:
            super().__init__(**kwargs)

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return ObjectTemplateType.BaseProgramGenerator

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            raise RuntimeError("class_name is None.")

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        code_file = cfg.get("code_file", "")
        render_data = cfg.get("render_data", None)
        return cls(code_file=get_template_path(code_file), render_data=render_data)

    @property
    def template_content(self) -> str:
        return self.base_template_content

    @template_content.setter
    def template_content(self, _template_content) -> None:
        self.base_template_content = _template_content

    @property
    def render_data(self) -> dict:
        return self.exec_data

    @render_data.setter
    def render_data(self, exec_data: dict) -> None:
        self.exec_data = exec_data

    @property
    def render_code(self) -> str:
        return self.exec_code

    @render_code.setter
    def render_code(self, exec_code) -> None:
        self.exec_code = exec_code


class QueryProgramGenerator(CodeGenerator):
    exec_code: Optional[str] = Field(
        default="", description="执行代码"
    )
    dreams_query_template_content: Optional[str] = Field(
        default="", description="模板内容",
    )
    exec_data: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="执行环境数据",
        alias="exec_data",
    )

    def __init__(self, query_code_file: str = None, render_data=None, **kwargs):

        # 检查kwargs中是否包含这些属性，如果包含就不执行文件读取
        if "exec_data" not in kwargs and "base_template_content" not in kwargs and "exec_code" not in kwargs:
            if render_data is None:
                render_data = {}
            # 读取模板文件
            with open(query_code_file, 'r') as dreams_query_template_file:
                dreams_query_template_content = dreams_query_template_file.read()

            super().__init__(exec_data=render_data, dreams_query_template_content=dreams_query_template_content)
        else:
            super().__init__(**kwargs)

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return ObjectTemplateType.QueryProgramGenerator

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            raise RuntimeError("class_name is None.")

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        query_code_file = cfg.get("query_code_file", "")
        render_data = cfg.get("render_data", None)
        return cls(query_code_file=get_template_path(query_code_file), render_data=render_data)

    @property
    def template_content(self) -> str:
        return self.dreams_query_template_content

    @template_content.setter
    def template_content(self, _template_content) -> None:
        self.dreams_query_template_content = _template_content

    @property
    def render_data(self) -> dict:
        return self.exec_data

    @render_data.setter
    def render_data(self, exec_data: dict) -> None:
        self.exec_data = exec_data

    @property
    def render_code(self) -> str:
        return self.exec_code

    @render_code.setter
    def render_code(self, exec_code) -> None:
        self.exec_code = exec_code


class AIProgramGenerator(CodeGenerator):
    exec_code: Optional[str] = Field(
        default="", description="执行代码"
    )
    ai_template_content: Optional[str] = Field(
        default="", description="模板内容",
    )
    exec_data: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="执行环境数据",
        alias="exec_data",
    )

    def __init__(self, ai_code_file: str = None, render_data=None, **kwargs):

        # 检查kwargs中是否包含这些属性，如果包含就不执行文件读取
        if "exec_data" not in kwargs and "base_template_content" not in kwargs and "exec_code" not in kwargs:
            if render_data is None:
                render_data = {}
            # 读取模板文件
            with open(ai_code_file, 'r') as ai_template_file:
                ai_template_content = ai_template_file.read()

            super().__init__(exec_data=render_data, ai_template_content=ai_template_content)
        else:
            super().__init__(**kwargs)

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return ObjectTemplateType.AIProgramGenerator

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            raise RuntimeError("class_name is None.")

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        ai_code_file = cfg.get("ai_code_file", "")
        render_data = cfg.get("render_data", None)
        return cls(ai_code_file=get_template_path(ai_code_file), render_data=render_data)

    @property
    def template_content(self) -> str:
        return self.ai_template_content

    @template_content.setter
    def template_content(self, _template_content) -> None:
        self.ai_template_content = _template_content

    @property
    def render_data(self) -> dict:
        return self.exec_data

    @render_data.setter
    def render_data(self, exec_data: dict) -> None:
        self.exec_data = exec_data

    @property
    def render_code(self) -> str:
        return self.exec_code

    @render_code.setter
    def render_code(self, exec_code) -> None:
        self.exec_code = exec_code


class EngineProgramGenerator(CodeGenerator):
    exec_code: Optional[str] = Field(
        default="", description="执行代码"
    )
    engine_template_content: Optional[str] = Field(
        default="", description="模板内容",
    )
    exec_data: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="执行环境数据",
        alias="exec_data",
    )

    def __init__(self, engine_code_file: str = None, render_data=None, **kwargs):

        # 检查kwargs中是否包含这些属性，如果包含就不执行文件读取
        if "exec_data" not in kwargs and "base_template_content" not in kwargs and "exec_code" not in kwargs:
            if render_data is None:
                render_data = {
                    'model_name': 'gpt-4',
                }
            model_name = render_data.get("model_name", None)
            if model_name is None:
                raise RuntimeError("model_name is None.")
            # 读取模板文件
            with open(engine_code_file, 'r') as engine_template_file:
                engine_template_content = engine_template_file.read()

            super().__init__(exec_data=render_data, engine_template_content=engine_template_content)
        else:
            super().__init__(**kwargs)

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return ObjectTemplateType.EngineProgramGenerator

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            raise RuntimeError("class_name is None.")

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        engine_code_file = cfg.get("engine_code_file", "")
        render_data = cfg.get("render_data", None)
        return cls(engine_code_file=get_template_path(engine_code_file), render_data=render_data)

    @property
    def template_content(self) -> str:
        return self.engine_template_content

    @template_content.setter
    def template_content(self, _template_content) -> None:
        self.engine_template_content = _template_content

    @property
    def render_data(self) -> dict:
        return self.exec_data

    @render_data.setter
    def render_data(self, exec_data: dict) -> None:
        self.exec_data = exec_data

    @property
    def render_code(self) -> str:
        return self.exec_code

    @render_code.setter
    def render_code(self, exec_code) -> None:
        self.exec_code = exec_code



import logging

import time
import os

logger = logging.getLogger(__name__)


class LoggerNameFilter(logging.Filter):
    def filter(self, record):
        # return record.name.startswith("loom_core") or record.name in "ERROR" or (
        #         record.name.startswith("uvicorn.error")
        #         and record.getMessage().startswith("Uvicorn running on")
        # )
        return True


def get_log_file(log_path: str, sub_dir: str):
    """
    sub_dir should contain a timestamp.
    """
    log_dir = os.path.join(log_path, sub_dir)
    # Here should be creating a new directory each time, so `exist_ok=False`
    os.makedirs(log_dir, exist_ok=False)
    return os.path.join(log_dir, "loom_core.log")


def get_config_dict(
        log_level: str, log_file_path: str, log_backup_count: int, log_max_bytes: int
) -> dict:
    # for windows, the path should be a raw string.
    log_file_path = (
        log_file_path.encode("unicode-escape").decode()
        if os.name == "nt"
        else log_file_path
    )
    log_level = log_level.upper()
    config_dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "formatter": {
                "format": (
                    "%(asctime)s %(name)-12s %(process)d %(levelname)-8s %(message)s"
                )
            },
        },
        "filters": {
            "logger_name_filter": {
                "()": __name__ + ".LoggerNameFilter",
            },
        },
        "handlers": {
            "stream_handler": {
                "class": "logging.StreamHandler",
                "formatter": "formatter",
                "level": log_level,
                # "stream": "ext://sys.stdout",
                # "filters": ["logger_name_filter"],
            },
            "file_handler": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "formatter",
                "level": log_level,
                "filename": log_file_path,
                "mode": "a",
                "maxBytes": log_max_bytes,
                "backupCount": log_backup_count,
                "encoding": "utf8",
            },
        },
        "loggers": {
            "loom_core": {
                "handlers": ["stream_handler", "file_handler"],
                "level": log_level,
                "propagate": False,
            }
        },
        "root": {
            "level": log_level,
            "handlers": ["stream_handler", "file_handler"],
        },
    }
    return config_dict


def get_timestamp_ms():
    t = time.time()
    return int(round(t * 1000))



"""IndexStructType class."""

from enum import Enum


class IndexStructType(str, Enum):
    """Index struct type. Identifier for a "type" of index.

    Attributes:
        DICT ("dict"): Faiss Vector Store Index. See
            :ref:`Ref-Indices-VectorStore`
            for more information on the faiss vector store index.

    """

    # TODO: refactor so these are properties on the base class

    CODE_GENERATOR = "code_generator"



"""Index registry."""

from typing import Dict, Type

from dreamsboard.engine.data_structs.data_structs import (
    IndexDict, IndexStruct,
)
from dreamsboard.engine.data_structs.struct_type import IndexStructType


INDEX_STRUCT_TYPE_TO_INDEX_STRUCT_CLASS: Dict[IndexStructType, Type[IndexStruct]] = {
    IndexStructType.CODE_GENERATOR: IndexDict,
}



from typing import List

from langchain.schema import (
    AIMessage,
    SystemMessage,
)
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable


# 创建一个代码执行器
class CodeExecutor:
    executor_code: str
    _messages: List[BaseMessage] = []
    _llm_runable: Runnable[LanguageModelInput, BaseMessage]
    _ai_message: AIMessage

    def __init__(self, executor_code: str, llm_runable: Runnable[LanguageModelInput, BaseMessage], messages: List[BaseMessage] = []):
        self.executor_code = executor_code
        self._llm_runable = llm_runable
        self._messages = messages

    def execute(self):
        # 创建变量字典
        variables = {
            'messages': self._messages,  # type: list[BaseMessage]
            'llm_runable': self._llm_runable,  # type: Runnable[LanguageModelInput, BaseMessage]
        }

        exec(self.executor_code, variables)

        self._llm_runable = variables.get('llm_runable', None)
        self._messages = variables.get('messages', None)

    def chat_run(self) -> AIMessage:
        # 创建变量字典
        variables = {
            'messages': self._messages,
            'llm_runable': self._llm_runable,
            'ai_message': None,
        }

        exec_code = 'ai_message = llm_runable.invoke(messages)\r\n'
        exec(exec_code, variables)
        self._ai_message = variables.get('ai_message', None)
        return self._ai_message
